"""ProtocolEngine-based InstrumentContext core implementation."""
from __future__ import annotations

from typing import Optional, TYPE_CHECKING, cast, Union
from opentrons.protocols.api_support.types import APIVersion

from opentrons.types import Location, Mount
from opentrons.hardware_control import SyncHardwareAPI
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.protocols.api_support.util import FlowRates, find_value_for_api_version
from opentrons.protocol_engine import (
    DeckPoint,
    DropTipWellLocation,
    DropTipWellOrigin,
    WellLocation,
    WellOrigin,
    WellOffset,
    AllNozzleLayoutConfiguration,
    SingleNozzleLayoutConfiguration,
    RowNozzleLayoutConfiguration,
    ColumnNozzleLayoutConfiguration,
    QuadrantNozzleLayoutConfiguration,
)
from opentrons.protocol_engine.types import (
    PRIMARY_NOZZLE_LITERAL,
    NozzleLayoutConfigurationType,
    AddressableOffsetVector,
)
from opentrons.protocol_engine.errors.exceptions import TipNotAttachedError
from opentrons.protocol_engine.clients import SyncClient as EngineClient
from opentrons.protocols.api_support.definitions import MAX_SUPPORTED_VERSION

from opentrons_shared_data.pipette.dev_types import PipetteNameType
from opentrons.protocol_api._nozzle_layout import NozzleLayout
from opentrons.hardware_control.nozzle_manager import NozzleConfigurationType
from . import deck_conflict

from ..instrument import AbstractInstrument
from .well import WellCore

from ..._trash_bin import TrashBin
from ..._waste_chute import WasteChute

if TYPE_CHECKING:
    from .protocol import ProtocolCore


_DISPENSE_VOLUME_VALIDATION_ADDED_IN = APIVersion(2, 17)


class InstrumentCore(AbstractInstrument[WellCore]):
    """Instrument API core using a ProtocolEngine.

    Args:
        pipette_id: ProtocolEngine ID of the loaded instrument.
    """

    def __init__(
        self,
        pipette_id: str,
        engine_client: EngineClient,
        sync_hardware_api: SyncHardwareAPI,
        protocol_core: ProtocolCore,
        default_movement_speed: float,
    ) -> None:
        self._pipette_id = pipette_id
        self._engine_client = engine_client
        self._sync_hardware_api = sync_hardware_api
        self._protocol_core = protocol_core

        # TODO(jbl 2022-11-03) flow_rates should not live in the cores, and should be moved to the protocol context
        #   along with other rate related refactors (for the hardware API)
        flow_rates = self._engine_client.state.pipettes.get_flow_rates(pipette_id)
        self._aspirate_flow_rate = find_value_for_api_version(
            MAX_SUPPORTED_VERSION, flow_rates.default_aspirate
        )
        self._dispense_flow_rate = find_value_for_api_version(
            MAX_SUPPORTED_VERSION, flow_rates.default_dispense
        )
        self._blow_out_flow_rate = find_value_for_api_version(
            MAX_SUPPORTED_VERSION, flow_rates.default_blow_out
        )
        self._flow_rates = FlowRates(self)

        self.set_default_speed(speed=default_movement_speed)

    @property
    def pipette_id(self) -> str:
        """The instrument's unique ProtocolEngine ID."""
        return self._pipette_id

    def get_default_speed(self) -> float:
        speed = self._engine_client.state.pipettes.get_movement_speed(
            pipette_id=self._pipette_id
        )
        assert speed is not None, "Pipette loading should have set a default speed."
        return speed

    def set_default_speed(self, speed: float) -> None:
        self._engine_client.set_pipette_movement_speed(
            pipette_id=self._pipette_id, speed=speed
        )

    def aspirate(
        self,
        location: Location,
        well_core: Optional[WellCore],
        volume: float,
        rate: float,
        flow_rate: float,
        in_place: bool,
    ) -> None:
        """Aspirate a given volume of liquid from the specified location.
        Args:
            volume: The volume of liquid to aspirate, in microliters.
            location: The exact location to aspirate from.
            well_core: The well to aspirate from, if applicable.
            rate: Not used in this core.
            flow_rate: The flow rate in µL/s to aspirate at.
            in_place: whether this is a in-place command.
        """
        if well_core is None:
            if not in_place:
                self._engine_client.move_to_coordinates(
                    pipette_id=self._pipette_id,
                    coordinates=DeckPoint(
                        x=location.point.x, y=location.point.y, z=location.point.z
                    ),
                    minimum_z_height=None,
                    force_direct=False,
                    speed=None,
                )

            self._engine_client.aspirate_in_place(
                pipette_id=self._pipette_id, volume=volume, flow_rate=flow_rate
            )

        else:
            well_name = well_core.get_name()
            labware_id = well_core.labware_id

            well_location = (
                self._engine_client.state.geometry.get_relative_well_location(
                    labware_id=labware_id,
                    well_name=well_name,
                    absolute_point=location.point,
                )
            )
            deck_conflict.check_safe_for_pipette_movement(
                engine_state=self._engine_client.state,
                pipette_id=self._pipette_id,
                labware_id=labware_id,
                well_name=well_name,
                well_location=well_location,
            )
            self._engine_client.aspirate(
                pipette_id=self._pipette_id,
                labware_id=labware_id,
                well_name=well_name,
                well_location=well_location,
                volume=volume,
                flow_rate=flow_rate,
            )

        self._protocol_core.set_last_location(location=location, mount=self.get_mount())

    def dispense(
        self,
        location: Union[Location, TrashBin, WasteChute],
        well_core: Optional[WellCore],
        volume: float,
        rate: float,
        flow_rate: float,
        in_place: bool,
        push_out: Optional[float],
    ) -> None:
        """Dispense a given volume of liquid into the specified location.
        Args:
            volume: The volume of liquid to dispense, in microliters.
            location: The exact location to dispense to.
            well_core: The well to dispense to, if applicable.
            rate: Not used in this core.
            flow_rate: The flow rate in µL/s to dispense at.
            in_place: whether this is a in-place command.
            push_out: The amount to push the plunger below bottom position.
        """
        if self._protocol_core.api_version < _DISPENSE_VOLUME_VALIDATION_ADDED_IN:
            # In older API versions, when you try to dispense more than you can,
            # it gets clamped.
            volume = min(volume, self.get_current_volume())
        else:
            # Newer API versions raise an error if you try to dispense more than
            # you can. Let the error come from Protocol Engine's validation.
            pass

        if well_core is None:
            if not in_place:
                if isinstance(location, (TrashBin, WasteChute)):
                    self._move_to_disposal_location(
                        disposal_location=location, force_direct=False, speed=None
                    )
                else:
                    self._engine_client.move_to_coordinates(
                        pipette_id=self._pipette_id,
                        coordinates=DeckPoint(
                            x=location.point.x, y=location.point.y, z=location.point.z
                        ),
                        minimum_z_height=None,
                        force_direct=False,
                        speed=None,
                    )

            self._engine_client.dispense_in_place(
                pipette_id=self._pipette_id,
                volume=volume,
                flow_rate=flow_rate,
                push_out=push_out,
            )
        else:
            if isinstance(location, (TrashBin, WasteChute)):
                raise ValueError("Trash Bin and Waste Chute have no Wells.")
            well_name = well_core.get_name()
            labware_id = well_core.labware_id

            well_location = (
                self._engine_client.state.geometry.get_relative_well_location(
                    labware_id=labware_id,
                    well_name=well_name,
                    absolute_point=location.point,
                )
            )
            deck_conflict.check_safe_for_pipette_movement(
                engine_state=self._engine_client.state,
                pipette_id=self._pipette_id,
                labware_id=labware_id,
                well_name=well_name,
                well_location=well_location,
            )
            self._engine_client.dispense(
                pipette_id=self._pipette_id,
                labware_id=labware_id,
                well_name=well_name,
                well_location=well_location,
                volume=volume,
                flow_rate=flow_rate,
                push_out=push_out,
            )

        if isinstance(location, (TrashBin, WasteChute)):
            self._protocol_core.set_last_location(location=None, mount=self.get_mount())
        else:
            self._protocol_core.set_last_location(
                location=location, mount=self.get_mount()
            )

    def blow_out(
        self,
        location: Union[Location, TrashBin, WasteChute],
        well_core: Optional[WellCore],
        in_place: bool,
    ) -> None:
        """Blow liquid out of the tip.

        Args:
            location: The location to blow out into.
            well_core: The well to blow out into.
            in_place: whether this is a in-place command.
        """
        flow_rate = self.get_blow_out_flow_rate(1.0)
        if well_core is None:
            if not in_place:
                if isinstance(location, (TrashBin, WasteChute)):
                    self._move_to_disposal_location(
                        disposal_location=location, force_direct=False, speed=None
                    )
                else:
                    self._engine_client.move_to_coordinates(
                        pipette_id=self._pipette_id,
                        coordinates=DeckPoint(
                            x=location.point.x, y=location.point.y, z=location.point.z
                        ),
                        force_direct=False,
                        minimum_z_height=None,
                        speed=None,
                    )

            self._engine_client.blow_out_in_place(
                pipette_id=self._pipette_id, flow_rate=flow_rate
            )
        else:
            if isinstance(location, (TrashBin, WasteChute)):
                raise ValueError("Trash Bin and Waste Chute have no Wells.")
            well_name = well_core.get_name()
            labware_id = well_core.labware_id

            well_location = (
                self._engine_client.state.geometry.get_relative_well_location(
                    labware_id=labware_id,
                    well_name=well_name,
                    absolute_point=location.point,
                )
            )
            deck_conflict.check_safe_for_pipette_movement(
                engine_state=self._engine_client.state,
                pipette_id=self._pipette_id,
                labware_id=labware_id,
                well_name=well_name,
                well_location=well_location,
            )
            self._engine_client.blow_out(
                pipette_id=self._pipette_id,
                labware_id=labware_id,
                well_name=well_name,
                well_location=well_location,
                # TODO(jbl 2022-11-07) PAPIv2 does not have an argument for rate and
                #   this also needs to be refactored along with other flow rate related issues
                flow_rate=flow_rate,
            )

        if isinstance(location, (TrashBin, WasteChute)):
            self._protocol_core.set_last_location(location=None, mount=self.get_mount())
        else:
            self._protocol_core.set_last_location(
                location=location, mount=self.get_mount()
            )

    def touch_tip(
        self,
        location: Location,
        well_core: WellCore,
        radius: float,
        z_offset: float,
        speed: float,
    ) -> None:
        """Touch pipette tip to edges of the well

        Args:
            location: Location moved to, only used for ProtocolCore location cache.
            well_core: The target well for touch tip.
            radius: Percentage modifier for well radius to touch.
            z_offset: Vertical offset for pipette tip during touch tip.
            speed: Speed for the touch tip movements.
        """
        well_name = well_core.get_name()
        labware_id = well_core.labware_id

        # Touch tip is always done from the top of the well.
        well_location = WellLocation(
            origin=WellOrigin.TOP, offset=WellOffset(x=0, y=0, z=z_offset)
        )
        deck_conflict.check_safe_for_pipette_movement(
            engine_state=self._engine_client.state,
            pipette_id=self._pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
        )
        self._engine_client.touch_tip(
            pipette_id=self._pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            radius=radius,
            speed=speed,
        )

        self._protocol_core.set_last_location(location=location, mount=self.get_mount())

    def pick_up_tip(
        self,
        location: Location,
        well_core: WellCore,
        presses: Optional[int],
        increment: Optional[float],
        prep_after: bool = True,
    ) -> None:
        """Move to and pick up a tip from a given well.

        Args:
            location: The location of the well we're picking up from.
                Used to calculate the relative well offset for the pick up command.
            well_core: The "well" to pick up from.
            presses: Customize the number of presses the pipette does.
            increment: Customize the movement "distance" of the pipette to press harder.
            prep_after: Not used by this core, pipette preparation will always happen.
        """
        assert (
            presses is None and increment is None
        ), "Tip pick-up with custom presses or increment deprecated"

        well_name = well_core.get_name()
        labware_id = well_core.labware_id

        well_location = self._engine_client.state.geometry.get_relative_well_location(
            labware_id=labware_id,
            well_name=well_name,
            absolute_point=location.point,
        )
        deck_conflict.check_safe_for_tip_pickup_and_return(
            engine_state=self._engine_client.state,
            pipette_id=self._pipette_id,
            labware_id=labware_id,
        )
        deck_conflict.check_safe_for_pipette_movement(
            engine_state=self._engine_client.state,
            pipette_id=self._pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
        )
        self._engine_client.pick_up_tip(
            pipette_id=self._pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
        )

        self._protocol_core.set_last_location(location=location, mount=self.get_mount())

    def drop_tip(
        self,
        location: Optional[Location],
        well_core: WellCore,
        home_after: Optional[bool],
        alternate_drop_location: Optional[bool] = False,
    ) -> None:
        """Move to and drop a tip into a given well.

        Args:
            location: The location of the well we're dropping tip into.
                Used to calculate the relative well offset for the drop command.
            well_core: The well we're dropping into
            home_after: Whether to home the pipette after the tip is dropped.
            alternate_drop_location: Whether to randomize the exact location to drop tip
                within the specified well.
        """
        well_name = well_core.get_name()
        labware_id = well_core.labware_id

        if location is not None:
            relative_well_location = (
                self._engine_client.state.geometry.get_relative_well_location(
                    labware_id=labware_id,
                    well_name=well_name,
                    absolute_point=location.point,
                )
            )

            well_location = DropTipWellLocation(
                origin=DropTipWellOrigin(relative_well_location.origin.value),
                offset=relative_well_location.offset,
            )
        else:
            well_location = DropTipWellLocation()

        if self._engine_client.state.labware.is_tiprack(labware_id):
            deck_conflict.check_safe_for_tip_pickup_and_return(
                engine_state=self._engine_client.state,
                pipette_id=self._pipette_id,
                labware_id=labware_id,
            )
        deck_conflict.check_safe_for_pipette_movement(
            engine_state=self._engine_client.state,
            pipette_id=self._pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
        )
        self._engine_client.drop_tip(
            pipette_id=self._pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            home_after=home_after,
            alternateDropLocation=alternate_drop_location,
        )

        self._protocol_core.set_last_location(location=location, mount=self.get_mount())

    def drop_tip_in_disposal_location(
        self, disposal_location: Union[TrashBin, WasteChute], home_after: Optional[bool]
    ) -> None:
        self._move_to_disposal_location(
            disposal_location,
            force_direct=False,
            speed=None,
            alternate_tip_drop=True,
        )
        self._drop_tip_in_place(home_after=home_after)
        self._protocol_core.set_last_location(location=None, mount=self.get_mount())

    def _move_to_disposal_location(
        self,
        disposal_location: Union[TrashBin, WasteChute],
        force_direct: bool,
        speed: Optional[float],
        alternate_tip_drop: bool = False,
    ) -> None:
        # TODO (nd, 2023-11-30): give appropriate offset when finalized
        # https://opentrons.atlassian.net/browse/RSS-391
        offset = AddressableOffsetVector(x=0, y=0, z=0)

        if isinstance(disposal_location, TrashBin):
            addressable_area_name = disposal_location._addressable_area_name
            self._engine_client.move_to_addressable_area_for_drop_tip(
                pipette_id=self._pipette_id,
                addressable_area_name=addressable_area_name,
                offset=offset,
                force_direct=force_direct,
                speed=speed,
                minimum_z_height=None,
                alternate_drop_location=alternate_tip_drop,
                ignore_tip_configuration=True,
            )

        if isinstance(disposal_location, WasteChute):
            num_channels = self.get_channels()
            addressable_area_name = {
                1: "1ChannelWasteChute",
                8: "8ChannelWasteChute",
                96: "96ChannelWasteChute",
            }[num_channels]

            self._engine_client.move_to_addressable_area(
                pipette_id=self._pipette_id,
                addressable_area_name=addressable_area_name,
                offset=offset,
                force_direct=force_direct,
                speed=speed,
                minimum_z_height=None,
            )

    def _drop_tip_in_place(self, home_after: Optional[bool]) -> None:
        self._engine_client.drop_tip_in_place(
            pipette_id=self._pipette_id,
            home_after=home_after,
        )

    def home(self) -> None:
        z_axis = self._engine_client.state.pipettes.get_z_axis(self._pipette_id)
        plunger_axis = self._engine_client.state.pipettes.get_plunger_axis(
            self._pipette_id
        )
        self._engine_client.home([z_axis, plunger_axis])

    def home_plunger(self) -> None:
        plunger_axis = self._engine_client.state.pipettes.get_plunger_axis(
            self._pipette_id
        )
        self._engine_client.home([plunger_axis])

    def move_to(
        self,
        location: Union[Location, TrashBin, WasteChute],
        well_core: Optional[WellCore],
        force_direct: bool,
        minimum_z_height: Optional[float],
        speed: Optional[float],
    ) -> None:
        if well_core is not None:
            if isinstance(location, (TrashBin, WasteChute)):
                raise ValueError("Trash Bin and Waste Chute have no Wells.")
            labware_id = well_core.labware_id
            well_name = well_core.get_name()
            well_location = (
                self._engine_client.state.geometry.get_relative_well_location(
                    labware_id=labware_id,
                    well_name=well_name,
                    absolute_point=location.point,
                )
            )

            self._engine_client.move_to_well(
                pipette_id=self._pipette_id,
                labware_id=labware_id,
                well_name=well_name,
                well_location=well_location,
                minimum_z_height=minimum_z_height,
                force_direct=force_direct,
                speed=speed,
            )
        else:
            if isinstance(location, (TrashBin, WasteChute)):
                self._move_to_disposal_location(
                    disposal_location=location, force_direct=force_direct, speed=speed
                )
            else:
                self._engine_client.move_to_coordinates(
                    pipette_id=self._pipette_id,
                    coordinates=DeckPoint(
                        x=location.point.x, y=location.point.y, z=location.point.z
                    ),
                    minimum_z_height=minimum_z_height,
                    force_direct=force_direct,
                    speed=speed,
                )
        if isinstance(location, (TrashBin, WasteChute)):
            self._protocol_core.set_last_location(location=None, mount=self.get_mount())
        else:
            self._protocol_core.set_last_location(
                location=location, mount=self.get_mount()
            )

    def get_mount(self) -> Mount:
        """Get the mount the pipette is attached to."""
        return self._engine_client.state.pipettes.get(
            self._pipette_id
        ).mount.to_hw_mount()

    def get_pipette_name(self) -> str:
        """Get the pipette's load name as a string.

        Will match the load name of the actually loaded pipette,
        which may differ from the requested load name.
        """
        # TODO (tz, 11-23-22): revert this change when merging
        # https://opentrons.atlassian.net/browse/RLIQ-251
        pipette = self._engine_client.state.pipettes.get(self._pipette_id)
        return (
            pipette.pipetteName.value
            if isinstance(pipette.pipetteName, PipetteNameType)
            else pipette.pipetteName
        )

    def get_model(self) -> str:
        return self._engine_client.state.pipettes.get_model_name(self._pipette_id)

    def get_display_name(self) -> str:
        return self._engine_client.state.pipettes.get_display_name(self._pipette_id)

    def get_min_volume(self) -> float:
        return self._engine_client.state.pipettes.get_minimum_volume(self._pipette_id)

    def get_max_volume(self) -> float:
        return self._engine_client.state.pipettes.get_maximum_volume(self._pipette_id)

    def get_working_volume(self) -> float:
        return self._engine_client.state.pipettes.get_working_volume(self._pipette_id)

    def get_current_volume(self) -> float:
        try:
            current_volume = self._engine_client.state.pipettes.get_aspirated_volume(
                self._pipette_id
            )
        except TipNotAttachedError:
            current_volume = None

        return current_volume or 0

    def get_available_volume(self) -> float:
        try:
            available_volume = self._engine_client.state.pipettes.get_available_volume(
                self._pipette_id
            )
        except TipNotAttachedError:
            available_volume = None

        return available_volume or 0

    def get_hardware_state(self) -> PipetteDict:
        """Get the current state of the pipette hardware as a dictionary."""
        return self._sync_hardware_api.get_attached_instrument(self.get_mount())  # type: ignore[no-any-return]

    def get_channels(self) -> int:
        return self._engine_client.state.tips.get_pipette_channels(self._pipette_id)

    def get_active_channels(self) -> int:
        return self._engine_client.state.tips.get_pipette_active_channels(
            self._pipette_id
        )

    def has_tip(self) -> bool:
        return (
            self._engine_client.state.pipettes.get_attached_tip(self._pipette_id)
            is not None
        )

    def get_return_height(self) -> float:
        return self._engine_client.state.pipettes.get_return_tip_scale(self._pipette_id)

    def get_flow_rate(self) -> FlowRates:
        return self._flow_rates

    def get_aspirate_flow_rate(self, rate: float = 1.0) -> float:
        return self._aspirate_flow_rate * rate

    def get_dispense_flow_rate(self, rate: float = 1.0) -> float:
        return self._dispense_flow_rate * rate

    def get_blow_out_flow_rate(self, rate: float = 1.0) -> float:
        return self._blow_out_flow_rate * rate

    def get_nozzle_configuration(self) -> NozzleConfigurationType:
        return self._engine_client.state.pipettes.get_nozzle_layout_type(
            self._pipette_id
        )

    def is_tip_tracking_available(self) -> bool:
        primary_nozzle = self._engine_client.state.pipettes.get_primary_nozzle(
            self._pipette_id
        )
        if self.get_nozzle_configuration() == NozzleConfigurationType.FULL:
            return True
        else:
            if self.get_channels() == 96:
                # SINGLE configuration with H12 nozzle is technically supported by the
                # current tip tracking implementation but we don't do any deck conflict
                # checks for it, so we won't provide full support for it yet.
                return (
                    self.get_nozzle_configuration() == NozzleConfigurationType.COLUMN
                    and primary_nozzle == "A12"
                )
            if self.get_channels() == 8:
                return (
                    self.get_nozzle_configuration() == NozzleConfigurationType.SINGLE
                    and primary_nozzle == "H1"
                )
        return False

    def set_flow_rate(
        self,
        aspirate: Optional[float] = None,
        dispense: Optional[float] = None,
        blow_out: Optional[float] = None,
    ) -> None:
        if aspirate is not None:
            assert aspirate > 0
            self._aspirate_flow_rate = aspirate
        if dispense is not None:
            assert dispense > 0
            self._dispense_flow_rate = dispense
        if blow_out is not None:
            assert blow_out > 0
            self._blow_out_flow_rate = blow_out

    def configure_for_volume(self, volume: float) -> None:
        self._engine_client.configure_for_volume(
            pipette_id=self._pipette_id, volume=volume
        )

    def prepare_to_aspirate(self) -> None:
        self._engine_client.prepare_to_aspirate(pipette_id=self._pipette_id)

    def configure_nozzle_layout(
        self,
        style: NozzleLayout,
        primary_nozzle: Optional[str],
        front_right_nozzle: Optional[str],
    ) -> None:
        if style == NozzleLayout.COLUMN:
            configuration_model: NozzleLayoutConfigurationType = (
                ColumnNozzleLayoutConfiguration(
                    primaryNozzle=cast(PRIMARY_NOZZLE_LITERAL, primary_nozzle)
                )
            )
        elif style == NozzleLayout.ROW:
            configuration_model = RowNozzleLayoutConfiguration(
                primaryNozzle=cast(PRIMARY_NOZZLE_LITERAL, primary_nozzle)
            )
        elif style == NozzleLayout.QUADRANT:
            assert front_right_nozzle is not None
            configuration_model = QuadrantNozzleLayoutConfiguration(
                primaryNozzle=cast(PRIMARY_NOZZLE_LITERAL, primary_nozzle),
                frontRightNozzle=front_right_nozzle,
            )
        elif style == NozzleLayout.SINGLE:
            configuration_model = SingleNozzleLayoutConfiguration(
                primaryNozzle=cast(PRIMARY_NOZZLE_LITERAL, primary_nozzle)
            )
        else:
            configuration_model = AllNozzleLayoutConfiguration()
        self._engine_client.configure_nozzle_layout(
            pipette_id=self._pipette_id, configuration_params=configuration_model
        )



"""Get tip presence command request, result and implementation models."""
from __future__ import annotations

from pydantic import Field, BaseModel
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from .pipetting_common import PipetteIdMixin
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

from ..types import TipPresenceStatus

if TYPE_CHECKING:
    from ..execution import TipHandler


GetTipPresenceCommandType = Literal["getTipPresence"]


class GetTipPresenceParams(PipetteIdMixin):
    """Payload required for a GetTipPresence command."""

    pass


class GetTipPresenceResult(BaseModel):
    """Result data from the execution of a GetTipPresence command."""

    status: TipPresenceStatus = Field(
        ...,
        description=(
            "Whether or not a tip is attached on the pipette. This only works on"
            " on FLEX because OT-2 pipettes do not possess tip presence sensors,"
            " hence, will always return TipPresenceStatus.UNKNOWN."
        ),
    )


class GetTipPresenceImplementation(
    AbstractCommandImpl[GetTipPresenceParams, GetTipPresenceResult]
):
    """GetTipPresence command implementation."""

    def __init__(
        self,
        tip_handler: TipHandler,
        **kwargs: object,
    ) -> None:
        self._tip_handler = tip_handler

    async def execute(self, params: GetTipPresenceParams) -> GetTipPresenceResult:
        """Verify if tip presence is as expected for the requested pipette."""
        pipette_id = params.pipetteId

        result = await self._tip_handler.get_tip_presence(
            pipette_id=pipette_id,
        )

        return GetTipPresenceResult(status=result)


class GetTipPresence(BaseCommand[GetTipPresenceParams, GetTipPresenceResult]):
    """GetTipPresence command model."""

    commandType: GetTipPresenceCommandType = "getTipPresence"
    params: GetTipPresenceParams
    result: Optional[GetTipPresenceResult]

    _ImplementationCls: Type[
        GetTipPresenceImplementation
    ] = GetTipPresenceImplementation


class GetTipPresenceCreate(BaseCommandCreate[GetTipPresenceParams]):
    """GetTipPresence command creation request model."""

    commandType: GetTipPresenceCommandType = "getTipPresence"
    params: GetTipPresenceParams

    _CommandCls: Type[GetTipPresence] = GetTipPresence



"""Load liquid command request, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel, Field
from typing import Optional, Type, Dict, TYPE_CHECKING
from typing_extensions import Literal

from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from ..state import StateView

LoadLiquidCommandType = Literal["loadLiquid"]


class LoadLiquidParams(BaseModel):
    """Payload required to load a liquid into a well."""

    liquidId: str = Field(
        ...,
        description="Unique identifier of the liquid to load.",
    )
    labwareId: str = Field(
        ...,
        description="Unique identifier of labware to load liquid into.",
    )
    volumeByWell: Dict[str, float] = Field(
        ...,
        description="Volume of liquid, in µL, loaded into each well by name, in this labware.",
    )


class LoadLiquidResult(BaseModel):
    """Result data from the execution of a LoadLiquid command."""

    pass


class LoadLiquidImplementation(AbstractCommandImpl[LoadLiquidParams, LoadLiquidResult]):
    """Load liquid command implementation."""

    def __init__(self, state_view: StateView, **kwargs: object) -> None:
        self._state_view = state_view

    async def execute(self, params: LoadLiquidParams) -> LoadLiquidResult:
        """Load data necessary for a liquid."""
        self._state_view.liquid.validate_liquid_id(params.liquidId)

        self._state_view.labware.validate_liquid_allowed_in_labware(
            labware_id=params.labwareId, wells=params.volumeByWell
        )

        return LoadLiquidResult()


class LoadLiquid(BaseCommand[LoadLiquidParams, LoadLiquidResult]):
    """Load liquid command resource model."""

    commandType: LoadLiquidCommandType = "loadLiquid"
    params: LoadLiquidParams
    result: Optional[LoadLiquidResult]

    _ImplementationCls: Type[LoadLiquidImplementation] = LoadLiquidImplementation


class LoadLiquidCreate(BaseCommandCreate[LoadLiquidParams]):
    """Load liquid command creation request."""

    commandType: LoadLiquidCommandType = "loadLiquid"
    params: LoadLiquidParams

    _CommandCls: Type[LoadLiquid] = LoadLiquid



from typing_extensions import Protocol


class Simulatable(Protocol):
    """Protocol specifying ability to simulate"""

    @property
    def is_simulator(self) -> bool:
        """`True` if this is a simulator; `False` otherwise."""
        ...



"""Shared code for managing pipette configuration and storage."""
from dataclasses import dataclass
import logging
from typing import (
    Dict,
    Optional,
    Tuple,
    Any,
    cast,
    List,
    TypeVar,
)
from typing_extensions import Final
import numpy
from opentrons_shared_data.pipette.dev_types import UlPerMmAction

from opentrons_shared_data.errors.exceptions import (
    CommandPreconditionViolated,
    CommandParameterLimitViolated,
    UnexpectedTipRemovalError,
    UnexpectedTipAttachError,
)
from opentrons_shared_data.pipette.pipette_definition import (
    liquid_class_for_volume_between_default_and_defaultlowvolume,
    PressFitPickUpTipConfiguration,
    CamActionPickUpTipConfiguration,
)

from opentrons import types as top_types
from opentrons.hardware_control.types import (
    CriticalPoint,
    HardwareAction,
    Axis,
    OT3Mount,
)
from opentrons.hardware_control.constants import (
    SHAKE_OFF_TIPS_SPEED,
    SHAKE_OFF_TIPS_PICKUP_DISTANCE,
    DROP_TIP_RELEASE_DISTANCE,
    SHAKE_OFF_TIPS_DROP_DISTANCE,
)

from opentrons.hardware_control.dev_types import PipetteDict
from .pipette import Pipette
from .instrument_calibration import (
    PipetteOffsetSummary,
    PipetteOffsetByPipetteMount,
    check_instrument_offset_reasonability,
)


MOD_LOG = logging.getLogger(__name__)

# TODO both pipette handlers should be combined once the pipette configurations
# are unified AND we separate out the concept of changing pipette state versus static state
HOME_POSITION: Final[float] = 230.15

MountType = TypeVar("MountType", top_types.Mount, OT3Mount)
InstrumentsByMount = Dict[MountType, Optional[Pipette]]
PipetteHandlingData = Tuple[Pipette, OT3Mount]


@dataclass(frozen=True)
class LiquidActionSpec:
    axis: Axis
    volume: float
    plunger_distance: float
    speed: float
    acceleration: float
    instr: Pipette
    current: float


@dataclass(frozen=True)
class TipActionMoveSpec:
    distance: float
    currents: Optional[Dict[Axis, float]]
    speed: Optional[
        float
    ]  # allow speed for a movement to default to its axes' speed settings


@dataclass(frozen=True)
class TipActionSpec:
    tip_action_moves: List[TipActionMoveSpec]
    shake_off_moves: List[Tuple[top_types.Point, Optional[float]]]
    z_distance_to_tiprack: Optional[float] = None
    ending_z_retract_distance: Optional[float] = None


class OT3PipetteHandler:
    IHP_LOG = MOD_LOG.getChild("InstrumentHandler")

    def __init__(self, attached_instruments: InstrumentsByMount[OT3Mount]):
        assert attached_instruments
        self._attached_instruments: InstrumentsByMount[OT3Mount] = attached_instruments
        self._ihp_log = self.__class__.IHP_LOG

    def reset_instrument(self, mount: Optional[OT3Mount] = None) -> None:
        """
        Reset the internal state of a pipette by its mount, without doing
        any lower level reconfiguration. This is useful to make sure that no
        settings changes from a protocol persist.

        :param mount: If specified, reset that mount. If not specified,
                      reset both
        """

        # need to have a reset function on the pipette
        def _reset(m: OT3Mount) -> None:
            self._ihp_log.info(f"Resetting configuration for {m}")
            p = self._attached_instruments[m]
            if not p:
                return
            p.reset_pipette_offset(OT3Mount.from_mount(m), to_default=False)
            p.reload_configurations()
            p.reset_state()

        if not mount:
            for m in type(list(self._attached_instruments.keys())[0]):
                _reset(m)
        else:
            _reset(mount)

    def get_instrument_offset(self, mount: OT3Mount) -> Optional[PipetteOffsetSummary]:
        """Get the specified pipette's offset."""
        assert mount != OT3Mount.GRIPPER, "Wrong mount type to fetch pipette offset"
        try:
            pipette = self.get_pipette(mount)
        except top_types.PipetteNotAttachedError:
            return None
        return self._return_augmented_offset_data(
            pipette, mount, pipette.pipette_offset
        )

    def reset_instrument_offset(self, mount: OT3Mount, to_default: bool) -> None:
        """
        Temporarily reset the pipette offset to default values.
        :param mount: Modify the given mount.
        """
        pipette = self.get_pipette(mount)
        pipette.reset_pipette_offset(mount, to_default)

    def save_instrument_offset(
        self, mount: OT3Mount, delta: top_types.Point
    ) -> PipetteOffsetSummary:
        """
        Save a new instrument offset the pipette offset to a particular value.
        :param mount: Modify the given mount.
        :param delta: The offset to set for the pipette.
        """
        pipette = self.get_pipette(mount)
        offset_data = pipette.save_pipette_offset(mount, delta)
        return self._return_augmented_offset_data(pipette, mount, offset_data)

    def _return_augmented_offset_data(
        self,
        pipette: Pipette,
        mount: OT3Mount,
        offset_data: PipetteOffsetByPipetteMount,
    ) -> PipetteOffsetSummary:
        if mount == OT3Mount.LEFT:
            other_pipette = self._attached_instruments.get(OT3Mount.RIGHT, None)
            if other_pipette:
                other_offset = other_pipette.pipette_offset.offset
            else:
                other_offset = top_types.Point(0, 0, 0)
            reasonability = check_instrument_offset_reasonability(
                offset_data.offset, other_offset
            )
        else:
            other_pipette = self._attached_instruments.get(OT3Mount.LEFT, None)
            if other_pipette:
                other_offset = other_pipette.pipette_offset.offset
            else:
                other_offset = top_types.Point(0, 0, 0)
            reasonability = check_instrument_offset_reasonability(
                other_offset, offset_data.offset
            )
        return PipetteOffsetSummary(
            offset=offset_data.offset,
            source=offset_data.source,
            status=offset_data.status,
            last_modified=offset_data.last_modified,
            reasonability_check_failures=reasonability,
        )

    # TODO(mc, 2022-01-11): change returned map value type to `Optional[PipetteDict]`
    # instead of potentially returning an empty dict
    # For compatibility purposes only right now. We should change this
    # as soon as we can modify the /pipettes endpoint.
    def get_attached_instruments(self) -> Dict[OT3Mount, PipetteDict]:
        """Get the status dicts of the cached attached instruments.

        Also available as :py:meth:`get_attached_instruments`.

        This returns a dictified version of the
        :py:class:`hardware_control.instruments.pipette.Pipette` as a dict keyed by
        the :py:class:`top_types.Mount` to which the pipette is attached.
        If no pipette is attached on a given mount, the mount key will
        still be present but will have the value ``None``.

        Note that this is only a query of a cached value; to actively scan
        for changes, use :py:meth:`cache_instruments`. This process deactivates
        the motors and should be used sparingly.
        """
        return {
            m: self.get_attached_instrument(m)
            for m in self._attached_instruments.keys()
        }

    # TODO(mc, 2022-01-11): change return type to `Optional[PipetteDict]` instead
    # of potentially returning an empty dict
    def get_attached_instrument(self, mount: OT3Mount) -> PipetteDict:
        # TODO (lc 12-05-2022) Kill this code ASAP
        instr = self._attached_instruments[mount]
        result: Dict[str, Any] = {}
        if instr:
            configs = [
                "name",
                "aspirate_flow_rate",
                "dispense_flow_rate",
                "pipette_id",
                "current_volume",
                "display_name",
                "tip_length",
                "model",
                "blow_out_flow_rate",
                "working_volume",
                "tip_overlap",
                "available_volume",
                "return_tip_height",
                "default_aspirate_flow_rates",
                "default_blow_out_flow_rates",
                "default_dispense_flow_rates",
                "back_compat_names",
                "supported_tips",
            ]

            instr_dict = instr.as_dict()
            # TODO (spp, 2021-08-27): Revisit this logic. Why do we need to build
            #  this dict newly every time? Any why only a few items are being updated?
            for key in configs:
                result[key] = instr_dict[key]

            result["current_nozzle_map"] = instr.nozzle_manager.current_configuration
            result["min_volume"] = instr.liquid_class.min_volume
            result["max_volume"] = instr.liquid_class.max_volume
            result["channels"] = instr._max_channels
            result["has_tip"] = instr.has_tip
            result["tip_length"] = instr.current_tip_length
            result["aspirate_speed"] = self.plunger_speed(
                instr, instr.aspirate_flow_rate, "aspirate"
            )
            result["dispense_speed"] = self.plunger_speed(
                instr, instr.dispense_flow_rate, "dispense"
            )
            result["blow_out_speed"] = self.plunger_speed(
                instr, instr.blow_out_flow_rate, "dispense"
            )
            result["ready_to_aspirate"] = instr.ready_to_aspirate

            result["default_blow_out_speeds"] = {
                alvl: self.plunger_speed(instr, fr, "blowout")
                for alvl, fr in instr.blow_out_flow_rates_lookup.items()
            }

            result["default_dispense_speeds"] = {
                alvl: self.plunger_speed(instr, fr, "dispense")
                for alvl, fr in instr.dispense_flow_rates_lookup.items()
            }
            result["default_aspirate_speeds"] = {
                alvl: self.plunger_speed(instr, fr, "aspirate")
                for alvl, fr in instr.aspirate_flow_rates_lookup.items()
            }
            result[
                "default_push_out_volume"
            ] = instr.active_tip_settings.default_push_out_volume
            result[
                "pipette_bounding_box_offsets"
            ] = instr.config.pipette_bounding_box_offsets
        return cast(PipetteDict, result)

    @property
    def attached_instruments(self) -> Dict[OT3Mount, PipetteDict]:
        return self.get_attached_instruments()

    @property
    def hardware_instruments(self) -> InstrumentsByMount[OT3Mount]:
        """Do not write new code that uses this."""
        return self._attached_instruments

    def set_current_tiprack_diameter(
        self, mount: OT3Mount, tiprack_diameter: float
    ) -> None:
        instr = self.get_pipette(mount)
        self._ihp_log.info(
            "Updating tip rack diameter on pipette mount: "
            f"{mount}, tip diameter: {tiprack_diameter} mm"
        )
        instr.current_tiprack_diameter = tiprack_diameter

    def set_working_volume(self, mount: OT3Mount, tip_volume: float) -> None:
        instr = self.get_pipette(mount)
        if not instr:
            raise top_types.PipetteNotAttachedError(
                "No pipette attached to {} mount".format(mount.name)
            )
        self._ihp_log.info(
            "Updating working volume on pipette mount:"
            f"{mount}, tip volume: {tip_volume} ul"
        )
        instr.working_volume = tip_volume

    def calibrate_plunger(
        self,
        mount: OT3Mount,
        top: Optional[float] = None,
        bottom: Optional[float] = None,
        blow_out: Optional[float] = None,
        drop_tip: Optional[float] = None,
    ) -> None:
        """
        Set calibration values for the pipette plunger.
        This can be called multiple times as the user sets each value,
        or you can set them all at once.
        :param top: Touching but not engaging the plunger.
        :param bottom: Must be above the pipette's physical hard-stop, while
        still leaving enough room for 'blow_out'
        :param blow_out: Plunger is pushed down enough to expel all liquids.
        :param drop_tip: Position that causes the tip to be released from the
        pipette
        """
        instr = self.get_pipette(mount)
        pos_dict: Dict[str, float] = {
            "top": instr.plunger_positions.top,
            "bottom": instr.plunger_positions.bottom,
            "blow_out": instr.plunger_positions.blow_out,
            "drop_tip": instr.plunger_positions.drop_tip,
        }
        if top is not None:
            pos_dict["top"] = top
        if bottom is not None:
            pos_dict["bottom"] = bottom
        if blow_out is not None:
            pos_dict["blow_out"] = blow_out
        if drop_tip is not None:
            pos_dict["drop_tip"] = drop_tip
        instr.update_config_item(pos_dict)

    def set_flow_rate(
        self,
        mount: OT3Mount,
        aspirate: Optional[float] = None,
        dispense: Optional[float] = None,
        blow_out: Optional[float] = None,
    ) -> None:
        this_pipette = self.get_pipette(mount)
        if aspirate:
            this_pipette.aspirate_flow_rate = aspirate
        if dispense:
            this_pipette.dispense_flow_rate = dispense
        if blow_out:
            this_pipette.blow_out_flow_rate = blow_out

    def set_pipette_speed(
        self,
        mount: OT3Mount,
        aspirate: Optional[float] = None,
        dispense: Optional[float] = None,
        blow_out: Optional[float] = None,
    ) -> None:
        this_pipette = self.get_pipette(mount)
        if aspirate:
            this_pipette.aspirate_flow_rate = self.plunger_flowrate(
                this_pipette, aspirate, "aspirate"
            )
        if dispense:
            this_pipette.dispense_flow_rate = self.plunger_flowrate(
                this_pipette, dispense, "dispense"
            )
        if blow_out:
            this_pipette.blow_out_flow_rate = self.plunger_flowrate(
                this_pipette, blow_out, "blowout"
            )

    def instrument_max_height(
        self,
        mount: OT3Mount,
        retract_distance: float,
        critical_point: Optional[CriticalPoint],
    ) -> float:
        """Return max achievable height of the attached instrument
        based on the current critical point
        """
        cp = self.critical_point_for(mount, critical_point)

        max_height = HOME_POSITION - retract_distance + cp.z

        return max_height

    async def reset(self) -> None:
        self._attached_instruments = {
            k: None for k in self._attached_instruments.keys()
        }

    async def update_nozzle_configuration(
        self,
        mount: MountType,
        back_left_nozzle: str,
        front_right_nozzle: str,
        starting_nozzle: Optional[str] = None,
    ) -> None:
        instr = self._attached_instruments[OT3Mount.from_mount(mount)]
        if instr:
            instr.update_nozzle_configuration(
                back_left_nozzle, front_right_nozzle, starting_nozzle
            )

    async def reset_nozzle_configuration(self, mount: OT3Mount) -> None:
        instr = self._attached_instruments[OT3Mount.from_mount(mount)]
        if instr:
            instr.reset_nozzle_configuration()

    async def add_tip(self, mount: OT3Mount, tip_length: float) -> None:
        instr = self._attached_instruments[mount]
        attached = self.attached_instruments
        instr_dict = attached[mount]
        if instr and not instr.has_tip:
            instr.add_tip(tip_length=tip_length)
            # TODO (spp, 2021-08-27): These items are being updated in a local copy
            #  of the PipetteDict, which gets thrown away. Fix this.
            instr_dict["has_tip"] = True
            instr_dict["tip_length"] = tip_length
        else:
            self._ihp_log.warning(
                "attach tip called while tip already attached to {instr}"
            )

    async def remove_tip(self, mount: OT3Mount) -> None:
        instr = self._attached_instruments[mount]
        attached = self.attached_instruments
        instr_dict = attached[mount]
        if instr and instr.has_tip:
            instr.remove_tip()
            # TODO (spp, 2021-08-27): These items are being updated in a local copy
            #  of the PipetteDict, which gets thrown away. Fix this.
            instr_dict["has_tip"] = False
            instr_dict["tip_length"] = 0.0
        else:
            self._ihp_log.warning("detach tip called with no tip")

    def critical_point_for(
        self, mount: OT3Mount, cp_override: Optional[CriticalPoint] = None
    ) -> top_types.Point:
        """Return the current critical point of the specified mount.

        The mount's critical point is the position of the mount itself, if no
        pipette is attached, or the pipette's critical point (which depends on
        tip status).

        If `cp_override` is specified, and that critical point actually exists,
        it will be used instead. Invalid `cp_override`s are ignored.
        """
        pip = self._attached_instruments[OT3Mount.from_mount(mount)]
        if pip is not None and cp_override != CriticalPoint.MOUNT:
            return pip.critical_point(cp_override)
        else:
            return top_types.Point(0, 0, 0)

    def ready_for_tip_action(
        self, target: Pipette, action: HardwareAction, mount: OT3Mount
    ) -> None:
        if not target.has_tip:
            raise UnexpectedTipRemovalError(str(action), target.name, mount.name)
        if (
            action == HardwareAction.ASPIRATE
            and target.current_volume == 0
            and not target.ready_to_aspirate
        ):
            raise RuntimeError("Pipette not ready to aspirate")
        self._ihp_log.debug(f"{action} on {target.name}")

    def plunger_position(
        self, instr: Pipette, ul: float, action: "UlPerMmAction"
    ) -> float:
        mm = ul / instr.ul_per_mm(ul, action)
        position = instr.plunger_positions.bottom - mm
        return round(position, 6)

    def plunger_speed(
        self, instr: Pipette, ul_per_s: float, action: "UlPerMmAction"
    ) -> float:
        mm_per_s = ul_per_s / instr.ul_per_mm(instr.working_volume, action)
        return round(mm_per_s, 6)

    def plunger_flowrate(
        self, instr: Pipette, mm_per_s: float, action: "UlPerMmAction"
    ) -> float:
        ul_per_s = mm_per_s * instr.ul_per_mm(instr.working_volume, action)
        return round(ul_per_s, 6)

    def plunger_acceleration(self, instr: Pipette, ul_per_s_per_s: float) -> float:
        # using nominal ul/mm, to make sure accelerations are always the same
        # regardless of volume being aspirated/dispensed
        mm_per_s_per_s = ul_per_s_per_s / instr.config.shaft_ul_per_mm
        return round(mm_per_s_per_s, 6)

    def plan_check_aspirate(
        self,
        mount: OT3Mount,
        volume: Optional[float],
        rate: float,
    ) -> Optional[LiquidActionSpec]:
        """Check preconditions for aspirate, parse args, and calculate positions.

        While the mechanics of issuing an aspirate move itself are left to child
        classes, determining things like aspiration volume from the allowed argument
        types is invariant between machines, and this method gathers that functionality.

        Coalesce
        - Optional volumes

        Check
        - Aspiration volumes compared to max and remaining

        Calculate
        - Plunger distances (possibly calling an overridden plunger_volume)
        """
        instrument = self.get_pipette(mount)
        self.ready_for_tip_action(instrument, HardwareAction.ASPIRATE, mount)
        if volume is None:
            self._ihp_log.debug(
                "No aspirate volume defined. Aspirating up to "
                "max_volume for the pipette"
            )
            asp_vol = instrument.available_volume
        else:
            asp_vol = volume

        if asp_vol == 0:
            return None

        assert instrument.ok_to_add_volume(
            asp_vol
        ), "Cannot aspirate more than pipette max volume"

        dist = self.plunger_position(
            instrument, instrument.current_volume + asp_vol, "aspirate"
        )
        speed = self.plunger_speed(
            instrument, instrument.aspirate_flow_rate * rate, "aspirate"
        )
        acceleration = self.plunger_acceleration(
            instrument, instrument.flow_acceleration
        )

        return LiquidActionSpec(
            axis=Axis.of_main_tool_actuator(mount),
            volume=asp_vol,
            plunger_distance=dist,
            speed=speed,
            acceleration=acceleration,
            instr=instrument,
            current=instrument.plunger_motor_current.run,
        )

    def plan_check_dispense(
        self,
        mount: OT3Mount,
        volume: Optional[float],
        rate: float,
        push_out: Optional[float],
    ) -> Optional[LiquidActionSpec]:
        """Check preconditions for dispense, parse args, and calculate positions.

        While the mechanics of issuing a dispense move itself are left to child
        classes, determining things like dispense volume from the allowed argument
        types is invariant between machines, and this method gathers that functionality.

        Coalesce
        - Optional volumes

        Check
        - Dispense volumes compared to max and remaining

        Calculate
        - Plunger distances (possibly calling an overridden plunger_volume)
        """

        instrument = self.get_pipette(mount)
        self.ready_for_tip_action(instrument, HardwareAction.DISPENSE, mount)

        if volume is None:
            disp_vol = instrument.current_volume
            self._ihp_log.debug(
                "No dispense volume specified. Dispensing all "
                "remaining liquid ({}uL) from pipette".format(disp_vol)
            )
        else:
            disp_vol = volume

        # Ensure we don't dispense more than the current volume.
        #
        # This clamping is inconsistent with plan_check_aspirate(), which asserts
        # that its input is in bounds instead of clamping it. This is to match a quirk
        # of the OT-2 version of this class. Protocol Engine does its own clamping,
        # so we don't expect this to trigger in practice.
        disp_vol = min(instrument.current_volume, disp_vol)
        is_full_dispense = numpy.isclose(instrument.current_volume - disp_vol, 0)

        if disp_vol == 0:
            return None

        if is_full_dispense:
            if push_out is None:
                push_out_ul = instrument.push_out_volume
            else:
                push_out_ul = push_out
        else:
            if push_out is not None and push_out != 0:
                raise CommandPreconditionViolated(
                    message="Cannot push_out on a dispense that does not leave the pipette empty",
                    detail={
                        "command": "dispense",
                        "remaining-volume": str(instrument.current_volume - disp_vol),
                    },
                )
            push_out_ul = 0

        push_out_dist_mm = push_out_ul / instrument.ul_per_mm(push_out_ul, "blowout")

        if not instrument.ok_to_push_out(push_out_dist_mm):
            raise CommandParameterLimitViolated(
                command_name="dispense",
                parameter_name="push_out",
                limit_statement="less than pipette max blowout volume",
                actual_value=str(push_out_ul),
            )

        dist = self.plunger_position(
            instrument, instrument.current_volume - disp_vol, "dispense"
        )
        speed = self.plunger_speed(
            instrument, instrument.dispense_flow_rate * rate, "dispense"
        )
        acceleration = self.plunger_acceleration(
            instrument, instrument.flow_acceleration
        )
        return LiquidActionSpec(
            axis=Axis.of_main_tool_actuator(mount),
            volume=disp_vol,
            plunger_distance=dist + push_out_dist_mm,
            speed=speed,
            acceleration=acceleration,
            instr=instrument,
            current=instrument.plunger_motor_current.run,
        )

    def plan_check_blow_out(
        self, mount: OT3Mount, volume: Optional[float] = None
    ) -> LiquidActionSpec:
        """Check preconditions and calculate values for blowout."""
        instrument = self.get_pipette(mount)
        speed = self.plunger_speed(instrument, instrument.blow_out_flow_rate, "blowout")
        acceleration = self.plunger_acceleration(
            instrument, instrument.flow_acceleration
        )
        max_distance = (
            instrument.plunger_positions.blow_out - instrument.plunger_positions.bottom
        )
        if volume is None:
            distance_mm = max_distance
        else:
            ul = volume
            distance_mm = ul / instrument.ul_per_mm(ul, "blowout")
            if distance_mm > max_distance:
                raise CommandParameterLimitViolated(
                    command_name="blow_out",
                    parameter_name="volume",
                    limit_statement="less than the available distance for the plunger to move",
                    actual_value=str(volume),
                )

        return LiquidActionSpec(
            axis=Axis.of_main_tool_actuator(mount),
            volume=0,
            plunger_distance=distance_mm,
            speed=speed,
            acceleration=acceleration,
            instr=instrument,
            current=instrument.plunger_motor_current.run,
        )

    @staticmethod
    def _build_tip_motor_moves(
        prep_move_dist: float,
        clamp_move_dist: float,
        prep_move_speed: float,
        clamp_move_speed: float,
        tip_motor_current: float,
        plunger_current: float,
    ) -> List[TipActionMoveSpec]:
        return [
            TipActionMoveSpec(
                distance=prep_move_dist,
                speed=prep_move_speed,
                currents={
                    Axis.P_L: plunger_current,
                    Axis.Q: tip_motor_current,
                },
            ),
            TipActionMoveSpec(
                distance=prep_move_dist + clamp_move_dist,
                speed=clamp_move_speed,
                currents={
                    Axis.P_L: plunger_current,
                    Axis.Q: tip_motor_current,
                },
            ),
        ]

    @staticmethod
    def _build_pickup_shakes(
        instrument: Pipette,
    ) -> List[Tuple[top_types.Point, Optional[float]]]:
        def build_one_shake() -> List[Tuple[top_types.Point, Optional[float]]]:
            shake_dist = float(SHAKE_OFF_TIPS_PICKUP_DISTANCE)
            shake_speed = float(SHAKE_OFF_TIPS_SPEED)
            return [
                (top_types.Point(-shake_dist, 0, 0), shake_speed),  # left
                (top_types.Point(2 * shake_dist, 0, 0), shake_speed),  # right
                (top_types.Point(-shake_dist, 0, 0), shake_speed),  # center
                (top_types.Point(0, -shake_dist, 0), shake_speed),  # front
                (top_types.Point(0, 2 * shake_dist, 0), shake_speed),  # back
                (top_types.Point(0, -shake_dist, 0), shake_speed),  # center
                (top_types.Point(0, 0, DROP_TIP_RELEASE_DISTANCE), None),  # up
            ]

        return []

    def plan_ht_pick_up_tip(self, tip_count: int) -> TipActionSpec:
        # Prechecks: ready for pickup tip and press/increment are valid
        mount = OT3Mount.LEFT
        instrument = self.get_pipette(mount)
        if instrument.has_tip:
            raise UnexpectedTipAttachError("pick_up_tip", instrument.name, mount.name)
        self._ihp_log.debug(f"Picking up tip on {mount.name}")

        pick_up_config = instrument.get_pick_up_configuration_for_tip_count(tip_count)
        if not isinstance(pick_up_config, CamActionPickUpTipConfiguration):
            raise CommandPreconditionViolated(
                f"Low-throughput pick up tip got wrong config for {instrument.name} on {mount.name}"
            )

        tip_motor_moves = self._build_tip_motor_moves(
            prep_move_dist=pick_up_config.prep_move_distance,
            clamp_move_dist=pick_up_config.distance,
            prep_move_speed=pick_up_config.prep_move_speed,
            clamp_move_speed=pick_up_config.speed,
            plunger_current=instrument.plunger_motor_current.run,
            tip_motor_current=pick_up_config.current_by_tip_count[tip_count],
        )

        return TipActionSpec(
            tip_action_moves=tip_motor_moves,
            shake_off_moves=[],
            z_distance_to_tiprack=(-1 * pick_up_config.connect_tiprack_distance_mm),
            ending_z_retract_distance=instrument.config.end_tip_action_retract_distance_mm,
        )

    def plan_lt_pick_up_tip(
        self,
        mount: OT3Mount,
        tip_count: int,
        presses: Optional[int],
        increment: Optional[float],
    ) -> TipActionSpec:
        # Prechecks: ready for pickup tip and press/increment are valid
        instrument = self.get_pipette(mount)
        if instrument.has_tip:
            raise UnexpectedTipAttachError("pick_up_tip", instrument.name, mount.name)
        self._ihp_log.debug(f"Picking up tip on {mount.name}")

        pick_up_config = instrument.get_pick_up_configuration_for_tip_count(tip_count)
        if not isinstance(pick_up_config, PressFitPickUpTipConfiguration):
            raise CommandPreconditionViolated(
                f"Low-throughput pick up tip got wrong config for {instrument.name} on {mount.name}"
            )
        if presses is None or presses < 0:
            checked_presses = pick_up_config.presses
        else:
            checked_presses = presses

        if not increment or increment < 0:
            check_incr = pick_up_config.increment
        else:
            check_incr = increment

        pick_up_speed = pick_up_config.speed

        def build_presses() -> List[TipActionMoveSpec]:
            # Press the nozzle into the tip <presses> number of times,
            # moving further by <increment> mm after each press
            press_moves = []
            for i in range(checked_presses):
                # move nozzle down into the tip
                press_dist = -1.0 * pick_up_config.distance + -1.0 * check_incr * i
                press_moves.append(
                    TipActionMoveSpec(
                        distance=press_dist,
                        speed=pick_up_speed,
                        currents={
                            Axis.by_mount(mount): pick_up_config.current_by_tip_count[
                                tip_count
                            ]
                        },
                    )
                )
                # move nozzle back up
                backup_dist = -press_dist
                press_moves.append(
                    TipActionMoveSpec(
                        distance=backup_dist,
                        speed=pick_up_speed,
                        currents=None,
                    )
                )
            return press_moves

        return TipActionSpec(
            tip_action_moves=build_presses(),
            shake_off_moves=self._build_pickup_shakes(instrument),
            ending_z_retract_distance=instrument.config.end_tip_action_retract_distance_mm,
        )

    @staticmethod
    def _shake_off_tips_drop(
        tiprack_diameter: float,
    ) -> List[Tuple[top_types.Point, Optional[float]]]:
        # tips don't always fall off, especially if resting against
        # tiprack or other tips below it. To ensure the tip has fallen
        # first, shake the pipette to dislodge partially-sealed tips,
        # then second, raise the pipette so loosened tips have room to fall
        shake_off_dist = SHAKE_OFF_TIPS_DROP_DISTANCE
        if tiprack_diameter > 0.0:
            shake_off_dist = min(shake_off_dist, tiprack_diameter / 4)
        shake_off_dist = max(shake_off_dist, 1.0)
        speed = SHAKE_OFF_TIPS_SPEED
        return [
            (top_types.Point(-shake_off_dist, 0, 0), speed),  # left
            (top_types.Point(2 * shake_off_dist, 0, 0), speed),  # right
            (top_types.Point(-shake_off_dist, 0, 0), speed),  # center
            (top_types.Point(0, 0, DROP_TIP_RELEASE_DISTANCE), None),  # top
        ]

    def plan_lt_drop_tip(
        self,
        mount: OT3Mount,
    ) -> TipActionSpec:
        instrument = self.get_pipette(mount)
        config = instrument.drop_configurations.plunger_eject
        if not config:
            raise CommandPreconditionViolated(
                f"No plunger-eject drop tip configurations for {instrument.name} on {mount.name}"
            )
        drop_seq = [
            TipActionMoveSpec(
                distance=instrument.plunger_positions.drop_tip,
                speed=config.speed,
                currents={
                    Axis.of_main_tool_actuator(mount): config.current,
                },
            ),
            TipActionMoveSpec(
                distance=instrument.plunger_positions.bottom,
                speed=None,
                currents={
                    Axis.of_main_tool_actuator(
                        mount
                    ): instrument.plunger_motor_current.run,
                },
            ),
        ]

        return TipActionSpec(
            tip_action_moves=drop_seq,
            shake_off_moves=[],
        )

    def plan_ht_drop_tip(self) -> TipActionSpec:
        mount = OT3Mount.LEFT
        instrument = self.get_pipette(mount)
        config = instrument.drop_configurations.cam_action
        if not config:
            raise CommandPreconditionViolated(
                f"No cam-action drop tip configurations for {instrument.name} on {mount.name}"
            )

        drop_seq = self._build_tip_motor_moves(
            prep_move_dist=config.prep_move_distance,
            clamp_move_dist=config.distance,
            prep_move_speed=config.prep_move_speed,
            clamp_move_speed=config.speed,
            plunger_current=instrument.plunger_motor_current.run,
            tip_motor_current=config.current,
        )

        return TipActionSpec(
            tip_action_moves=drop_seq,
            shake_off_moves=[],
        )

    def has_pipette(self, mount: OT3Mount) -> bool:
        return bool(self._attached_instruments[mount])

    def get_pipette(self, mount: OT3Mount) -> Pipette:
        pip = self._attached_instruments[mount]
        if not pip:
            raise top_types.PipetteNotAttachedError(
                f"No pipette attached to {mount.name} mount"
            )
        return pip

    async def set_liquid_class(self, mount: OT3Mount, liquid_class: str) -> None:
        pip = self.get_pipette(mount)
        pip.set_liquid_class_by_name(liquid_class)

    async def configure_for_volume(self, mount: OT3Mount, volume: float) -> None:
        pip = self.get_pipette(mount)
        if pip.current_volume > 0:
            # Switching liquid classes can't happen when there's already liquid
            return
        new_class_name = liquid_class_for_volume_between_default_and_defaultlowvolume(
            volume,
            pip.liquid_class_name,
            pip.config.liquid_properties,
        )
        pip.set_liquid_class_by_name(new_class_name.name)

    def get_tip_sensor_count(self, mount: OT3Mount) -> int:
        if not self.has_pipette(mount):
            return 0
        return self.get_pipette(mount).tip_presence_responses



import asyncio
from typing import Callable, Dict, Tuple
from typing_extensions import Protocol
from opentrons.hardware_control.types import BoardRevision, DoorState
from .types import GPIOPin


class GPIODriverLike(Protocol):
    """Interface for the GPIO drivers"""

    def __init__(self, chip_name: str) -> None:
        ...

    @property
    def chip(self) -> str:
        ...

    @property
    def lines(self) -> Dict[str, int]:
        ...

    @property
    def board_rev(self) -> BoardRevision:
        ...

    @board_rev.setter
    def board_rev(self, boardrev: BoardRevision) -> None:
        ...

    async def setup(self) -> None:
        ...

    def config_by_board_rev(self) -> None:
        ...

    def set_high(self, output_pin: GPIOPin) -> None:
        ...

    def set_low(self, output_pin: GPIOPin) -> None:
        ...

    def set_button_light(
        self, red: bool = False, green: bool = False, blue: bool = False
    ) -> None:
        ...

    def set_rail_lights(self, on: bool = True) -> None:
        ...

    def set_reset_pin(self, on: bool = True) -> None:
        ...

    def set_isp_pin(self, on: bool = True) -> None:
        ...

    def set_halt_pin(self, on: bool = True) -> None:
        ...

    def get_button_light(self) -> Tuple[bool, bool, bool]:
        ...

    def get_rail_lights(self) -> bool:
        ...

    def read_button(self) -> bool:
        ...

    def read_window_switches(self) -> bool:
        ...

    def read_top_window_switch(self) -> bool:
        ...

    def read_front_door_switch(self) -> bool:
        ...

    def read_revision_bits(self) -> Tuple[bool, bool]:
        ...

    def get_door_state(self) -> DoorState:
        ...

    def start_door_switch_watcher(
        self,
        loop: asyncio.AbstractEventLoop,
        update_door_state: Callable[[DoorState], None],
    ) -> None:
        ...

    def release_line(self, pin: GPIOPin) -> None:
        ...

    def stop_door_switch_watcher(self, loop: asyncio.AbstractEventLoop) -> None:
        ...



import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .dev_types import GPIODriverLike

MODULE_LOG = logging.getLogger(__name__)


class RevisionPinsError(Exception):
    pass


def build_gpio_chardev(chip_name: str) -> "GPIODriverLike":
    try:
        from .gpio import GPIOCharDev

        return GPIOCharDev(chip_name)
    except (ImportError, OSError):
        MODULE_LOG.warning(
            "Failed to initialize character device, will not "
            "be able to control gpios (lights, button, smoothie"
            "kill, smoothie reset). Only one connection can be"
            " made to the gpios at a time. "
            "If you need to control gpios, first stop the robot "
            "server with systemctl stop opentrons-robot-server. "
            "Until you restart the server with systemctl start "
            "opentrons-robot-server, you will be unable to "
            "control the robot using the Opentrons app."
        )
        from .gpio_simulator import SimulatingGPIOCharDev

        return SimulatingGPIOCharDev(chip_name)



import asyncio
import logging
import os
from pathlib import Path
from glob import glob
from typing import Any, AsyncGenerator, Dict, Tuple, Optional, Union
from .types import UpdateError
from .mod_abc import AbstractModule
from opentrons.hardware_control.threaded_async_lock import ThreadedAsyncLock
from contextlib import asynccontextmanager

log = logging.getLogger(__name__)

_update_transition_lock = ThreadedAsyncLock()


@asynccontextmanager
async def protect_update_transition() -> AsyncGenerator[None, None]:
    async with _update_transition_lock.lock():
        yield


async def update_firmware(
    module: AbstractModule,
    firmware_file: Union[str, Path],
    loop: Optional[asyncio.AbstractEventLoop],
) -> None:
    """Apply update of given firmware file to given module.

    raises an UpdateError with the reason for the failure.
    """
    async with protect_update_transition():
        flash_port_or_dfu_serial = await module.prep_for_update()
        kwargs: Dict[str, Any] = {
            "stdout": asyncio.subprocess.PIPE,
            "stderr": asyncio.subprocess.PIPE,
        }
        successful, res = await module.bootloader()(
            flash_port_or_dfu_serial, str(firmware_file), kwargs
        )
        if not successful:
            log.info(f"Bootloader reponse: {res}")
            raise UpdateError(res)


async def find_bootloader_port() -> str:
    """
    Finds the port of an Opentrons Module that has entered its bootloader.
    The bootloader port shows up as 'ot_module_(avrdude|samba)_bootloader'
    on the pi; return found port.
    """

    for attempt in range(3):
        bootloader_ports = glob("/dev/ot_module_*_bootloader*")
        if bootloader_ports:
            if len(bootloader_ports) == 1:
                log.info(f"Found bootloader at port {bootloader_ports[0]}")
                return bootloader_ports[0]
            elif len(bootloader_ports) > 1:
                raise OSError("Multiple new bootloader ports" "found on mode switch")
        await asyncio.sleep(2)
    raise Exception("No ot_module bootloaders found in /dev. Try again")


async def find_dfu_device(pid: str, expected_device_count: int) -> str:
    """
    Find the dfu device and return its serial number (separate from module serial).

    Args:
        - pid: The USB Product ID of the device
        - expected_device_count: The expected number of "devices" for dfu-util
        to find for this PID. This is necessary because most STM32 MCU's
        will enumerate with multiple DFU devices, representing the
        separate programmable memory regions on the device. If more than
        this many devices are found, it is assumed that either the wrong
        module is in DFU mode *or* multiple modules are in DFU mode.
    """
    retries = 5
    log.info(f"Searching for a dfu device with PID {pid}")
    while retries != 0:
        retries -= 1
        await asyncio.sleep(1)
        proc = await asyncio.create_subprocess_exec(
            "dfu-util",
            "-l",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        await proc.wait()
        stdout, stderr = await proc.communicate()

        if stdout is None and stderr is None:
            continue
        if stderr:
            raise RuntimeError(f"Error finding dfu device: {stderr.decode()}")

        result = stdout.decode()
        if pid not in result:
            # It could take a few seconds for the device to show up
            continue
        devices_found = 0
        for line in result.splitlines():
            if pid in line:
                log.info(f"Found device with PID {pid}")
                devices_found += 1
                serial = line[(line.find("serial=") + 7) :]
        if devices_found == expected_device_count:
            # Heater-Shaker has 2 unique endpoints, Thermocycler has 3
            return serial
        elif devices_found > expected_device_count:
            raise OSError("Multiple new bootloader devices" "found on mode switch")

    raise RuntimeError(
        "Could not update firmware via dfu. Possible issues- dfu-util"
        " not working or specified dfu device not found"
    )


async def upload_via_avrdude(
    port: str, firmware_file_path: str, kwargs: Dict[str, Any]
) -> Tuple[bool, str]:
    """
    Run firmware upload command for hardware module with avrdude bootloader.

    Returns tuple of success boolean and message from bootloader.
    """
    # avrdude_options
    PART_NO = "atmega32u4"
    PROGRAMMER_ID = "avr109"
    BAUDRATE = "57600"

    config_file_path = Path("/etc/avrdude.conf")
    proc = await asyncio.create_subprocess_exec(
        "avrdude",
        "-C{}".format(config_file_path),
        "-v",
        "-p{}".format(PART_NO),
        "-c{}".format(PROGRAMMER_ID),
        "-P{}".format(port),
        "-b{}".format(BAUDRATE),
        "-D",
        "-Uflash:w:{}:i".format(firmware_file_path),
        **kwargs,
    )
    await proc.wait()

    _result = await proc.communicate()
    result = _result[1].decode()
    avrdude_res = _format_avrdude_response(result)
    if avrdude_res[0]:
        log.debug(result)
    else:
        log.error(
            "Failed to update module firmware for {}: {}".format(port, avrdude_res[1])
        )
    return avrdude_res


def _format_avrdude_response(raw_response: str) -> Tuple[bool, str]:
    avrdude_log = ""
    for line in raw_response.splitlines():
        if "avrdude:" in line and line != raw_response.splitlines()[1]:
            avrdude_log += line.lstrip("avrdude:") + ".."
            if "flash verified" in line:
                return True, line.lstrip("avrdude: ")
    return False, avrdude_log


async def upload_via_bossa(
    port: str, firmware_file_path: str, kwargs: Dict[str, Any]
) -> Tuple[bool, str]:
    """
    Run firmware upload command for hardware module with SAMBA bootloader.

    Returns tuple of success boolean and message from bootloader.
    """
    # bossac -p/dev/ttyACM1 -e -w -v -R --offset=0x2000
    #   modules/thermo-cycler/production/firmware/thermo-cycler-arduino.ino.bin
    # NOTE: bossac cannot traverse symlinks to port,
    # so we resolve to real path
    resolved_symlink = os.path.realpath(port)
    log.info(
        f"device at symlinked port: {port} " f"resolved to path: {resolved_symlink}"
    )
    bossa_args = [
        "bossac",
        f"-p{resolved_symlink}",
        "-e",
        "-w",
        "-v",
        "-R",
        "--offset=0x2000",
        f"{firmware_file_path}",
    ]

    proc = await asyncio.create_subprocess_exec(*bossa_args, **kwargs)
    stdout, stderr = await proc.communicate()
    res = stdout.decode()
    if "Verify successful" in res:
        log.debug(res)
        return True, res
    elif stderr:
        log.error(f"Failed to update module firmware for {port}: {res}")
        log.error(f"Error given: {stderr.decode()}")
        return False, res
    return False, ""


async def upload_via_dfu(
    dfu_serial: str, firmware_file_path: str, kwargs: Dict[str, Any]
) -> Tuple[bool, str]:
    """Run firmware upload command for DFU.

    Unlike other firmware upload methods, this one doesn't take a `port` argument since
    the module isn't recognized as a cdc device in dfu mode and hence doesn't get
    a port. The firmware upload utility, dfu-util, looks for the specific module
    by searching for available dfu devices. Since we check beforehand that only one
    dfu device is available during the upload process, this check is sufficient for us.

    In the future, if we want to make sure that the dfu device available is in fact
    the one we seek, then we can ask dfu-util to check for available dfu devices with
    a specific serial number (unrelated to Opentrons' module serial numbers).
    Hence, this method takes a `dfu_serial` argument instead.

    Returns tuple of success boolean and message from bootloader
    """
    log.info("Starting firmware upload via dfu util")
    dfu_args = [
        "dfu-util",
        "-a 0",
        "-s 0x08000000:leave",
        f"-D{firmware_file_path}",
        "-R",
    ]
    proc = await asyncio.create_subprocess_exec(*dfu_args, **kwargs)
    stdout, stderr = await proc.communicate()
    res = stdout.decode()

    if "File downloaded successfully" in res:
        log.debug(res)
        log.info("Firmware upload successful")
        return True, res
    else:
        log.error(
            f"Failed to update module firmware for {dfu_serial}. "
            # It isn't easy to decipher the issue from stderror alone
            f"stdout: {res} \n"
            f"stderr: {stderr.decode()}"
        )
        return False, res



"""Generate a JSON schema against which all create commands statically validate."""
import json
import pydantic
import argparse
import sys
from opentrons.protocol_engine.commands.command_unions import CommandCreate


class CreateCommandUnion(pydantic.BaseModel):
    """Model that validates a union of all CommandCreate models."""

    __root__: CommandCreate


def generate_command_schema(version: str) -> str:
    """Generate a JSON Schema that all valid create commands can validate against."""
    raw_json_schema = CreateCommandUnion.schema_json()
    schema_as_dict = json.loads(raw_json_schema)
    schema_as_dict["$id"] = f"opentronsCommandSchemaV{version}"
    schema_as_dict["$schema"] = "http://json-schema.org/draft-07/schema#"
    return json.dumps(schema_as_dict, indent=2)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="generate_command_schema",
        description="Generate A JSON-schema of all possible Create-Commands accepted by the current Protocol Engine",
    )
    parser.add_argument(
        "version",
        type=str,
        help="The command schema version. This is a single integer (e.g. 7) that will be used to name the generated schema file",
    )
    args = parser.parse_args()
    print(generate_command_schema(args.version))

    sys.exit()

__all__ = ["generate_command_schema"]



"""Protocol engine commands sub-state."""
from __future__ import annotations
from collections import OrderedDict
from enum import Enum
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, List, Optional, Union

from opentrons_shared_data.errors import EnumeratedError, ErrorCodes, PythonException

from opentrons.ordered_set import OrderedSet

from opentrons.hardware_control.types import DoorState

from ..actions import (
    Action,
    QueueCommandAction,
    UpdateCommandAction,
    FailCommandAction,
    PlayAction,
    PauseAction,
    StopAction,
    FinishAction,
    HardwareStoppedAction,
    DoorChangeAction,
)

from ..commands import Command, CommandStatus, CommandIntent
from ..errors import (
    CommandDoesNotExistError,
    RunStoppedError,
    ErrorOccurrence,
    RobotDoorOpenError,
    SetupCommandNotAllowedError,
    PauseNotAllowedError,
    UnexpectedProtocolError,
    ProtocolCommandFailedError,
)
from ..types import EngineStatus
from .abstract_store import HasState, HandlesActions
from .config import Config


class QueueStatus(str, Enum):
    """Execution status of the command queue.

    Properties:
        SETUP: The engine has been created, but the run has not yet started.
            New protocol commands may be enqueued but will wait to execute.
            New setup commands may be enqueued and will execute immediately.
        RUNNING: The queue is running though protocol commands.
            New protocol commands may be enqueued and will execute immediately.
            New setup commands may not be enqueued.
        PAUSED: Execution of protocol commands has been paused.
            New protocol commands may be enqueued but wait to execute.
            New setup commands may not be enqueued.
    """

    SETUP = "setup"
    RUNNING = "running"
    PAUSED = "paused"


class RunResult(str, Enum):
    """Result of the run."""

    SUCCEEDED = "succeeded"
    FAILED = "failed"
    STOPPED = "stopped"


@dataclass(frozen=True)
class CommandSlice:
    """A subset of all commands in state."""

    commands: List[Command]
    cursor: int
    total_length: int


@dataclass(frozen=True)
class CurrentCommand:
    """The "current" command's ID and index in the overall commands list."""

    command_id: str
    command_key: str
    created_at: datetime
    index: int


@dataclass(frozen=True)
class CommandEntry:
    """An command entry in state, including its index in the list."""

    command: Command
    index: int


@dataclass
class CommandState:
    """State of all protocol engine command resources."""

    all_command_ids: List[str]
    """All command IDs, in insertion order."""

    queued_command_ids: OrderedSet[str]
    """The IDs of queued commands, in FIFO order"""

    queued_setup_command_ids: OrderedSet[str]
    """The IDs of queued setup commands, in FIFO order"""

    running_command_id: Optional[str]
    """The ID of the currently running command, if any"""

    commands_by_id: Dict[str, CommandEntry]
    """All command resources, in insertion order, mapped by their unique IDs."""

    queue_status: QueueStatus
    """Whether the engine is currently pulling new commands off the queue to execute.

    A command may still be executing, and the robot may still be in motion,
    even if INACTIVE.
    """

    run_started_at: Optional[datetime]
    """The time the run was started.

    Set when the first `PlayAction` is dispatched.
    """

    run_completed_at: Optional[datetime]
    """The time the run has completed.

    Set when 'HardwareStoppedAction' is dispatched.
    """

    is_door_blocking: bool
    """Whether the door is open when enable_door_safety_switch feature flag is ON."""

    run_result: Optional[RunResult]
    """Whether the run is done and succeeded, failed, or stopped.

    This doesn't include the post-run finish steps (homing and dropping tips).

    Once set, this status is immutable.
    """

    run_error: Optional[ErrorOccurrence]
    """The run's fatal error occurrence, if there was one.

    Individual command errors, which may or may not be fatal,
    are stored on the individual commands themselves.
    """

    failed_command: Optional[CommandEntry]
    """The command, if any, that made the run fail and the index in the command list."""

    finish_error: Optional[ErrorOccurrence]
    """The error that happened during the post-run finish steps (homing & dropping tips), if any."""

    latest_command_hash: Optional[str]
    """The latest hash value received in a QueueCommandAction.

    This value can be used to generate future hashes.
    """

    stopped_by_estop: bool
    """If this is set to True, the engine was stopped by an estop event."""


class CommandStore(HasState[CommandState], HandlesActions):
    """Command state container."""

    _state: CommandState

    def __init__(
        self,
        *,
        config: Config,
        is_door_open: bool,
    ) -> None:
        """Initialize a CommandStore and its state."""
        self._config = config
        self._state = CommandState(
            queue_status=QueueStatus.SETUP,
            is_door_blocking=is_door_open and config.block_on_door_open,
            run_result=None,
            running_command_id=None,
            all_command_ids=[],
            queued_command_ids=OrderedSet(),
            queued_setup_command_ids=OrderedSet(),
            commands_by_id=OrderedDict(),
            run_error=None,
            finish_error=None,
            failed_command=None,
            run_completed_at=None,
            run_started_at=None,
            latest_command_hash=None,
            stopped_by_estop=False,
        )

    def handle_action(self, action: Action) -> None:  # noqa: C901
        """Modify state in reaction to an action."""
        if isinstance(action, QueueCommandAction):
            assert action.command_id not in self._state.commands_by_id

            # TODO(mc, 2021-06-22): mypy has trouble with this automatic
            # request > command mapping, figure out how to type precisely
            # (or wait for a future mypy version that can figure it out).
            # For now, unit tests cover mapping every request type
            queued_command = action.request._CommandCls.construct(
                id=action.command_id,
                key=(
                    action.request.key
                    if action.request.key is not None
                    else (action.request_hash or action.command_id)
                ),
                createdAt=action.created_at,
                params=action.request.params,  # type: ignore[arg-type]
                intent=action.request.intent,
                status=CommandStatus.QUEUED,
            )

            next_index = len(self._state.all_command_ids)
            self._state.all_command_ids.append(action.command_id)
            self._state.commands_by_id[queued_command.id] = CommandEntry(
                index=next_index,
                command=queued_command,
            )

            if action.request.intent == CommandIntent.SETUP:
                self._state.queued_setup_command_ids.add(queued_command.id)
            else:
                self._state.queued_command_ids.add(queued_command.id)

            if action.request_hash is not None:
                self._state.latest_command_hash = action.request_hash

        # TODO(mc, 2021-12-28): replace "UpdateCommandAction" with explicit
        # state change actions (e.g. RunCommandAction, SucceedCommandAction)
        # to make a command's queue transition logic easier to follow
        elif isinstance(action, UpdateCommandAction):
            command = action.command
            prev_entry = self._state.commands_by_id.get(command.id)

            if prev_entry is None:
                index = len(self._state.all_command_ids)
                self._state.all_command_ids.append(command.id)
                self._state.commands_by_id[command.id] = CommandEntry(
                    index=index,
                    command=command,
                )
            else:
                self._state.commands_by_id[command.id] = CommandEntry(
                    index=prev_entry.index,
                    command=command,
                )

            self._state.queued_command_ids.discard(command.id)
            self._state.queued_setup_command_ids.discard(command.id)

            if command.status == CommandStatus.RUNNING:
                self._state.running_command_id = command.id
            elif self._state.running_command_id == command.id:
                self._state.running_command_id = None

        elif isinstance(action, FailCommandAction):
            error_occurrence = ErrorOccurrence.from_failed(
                id=action.error_id,
                createdAt=action.failed_at,
                error=action.error,
            )
            prev_entry = self._state.commands_by_id[action.command_id]
            self._state.commands_by_id[action.command_id] = CommandEntry(
                index=prev_entry.index,
                # TODO(mc, 2022-06-06): add new "cancelled" status or similar
                # and don't set `completedAt` in commands other than the
                # specific one that failed
                command=prev_entry.command.copy(
                    update={
                        "error": error_occurrence,
                        "completedAt": action.failed_at,
                        "status": CommandStatus.FAILED,
                    }
                ),
            )

            self._state.failed_command = self._state.commands_by_id[action.command_id]
            if prev_entry.command.intent == CommandIntent.SETUP:
                other_command_ids_to_fail = [
                    *[i for i in self._state.queued_setup_command_ids],
                ]
                self._state.queued_setup_command_ids.clear()
            else:
                other_command_ids_to_fail = [
                    *[i for i in self._state.queued_command_ids],
                ]
                self._state.queued_command_ids.clear()

            for command_id in other_command_ids_to_fail:
                prev_entry = self._state.commands_by_id[command_id]

                self._state.commands_by_id[command_id] = CommandEntry(
                    index=prev_entry.index,
                    command=prev_entry.command.copy(
                        update={
                            "completedAt": action.failed_at,
                            "status": CommandStatus.FAILED,
                        }
                    ),
                )

            if self._state.running_command_id == action.command_id:
                self._state.running_command_id = None

        elif isinstance(action, PlayAction):
            if not self._state.run_result:
                self._state.run_started_at = (
                    self._state.run_started_at or action.requested_at
                )
                if self._state.is_door_blocking:
                    # Always inactivate queue when door is blocking
                    self._state.queue_status = QueueStatus.PAUSED
                else:
                    self._state.queue_status = QueueStatus.RUNNING

        elif isinstance(action, PauseAction):
            self._state.queue_status = QueueStatus.PAUSED

        elif isinstance(action, StopAction):
            if not self._state.run_result:
                self._state.queue_status = QueueStatus.PAUSED
                self._state.run_result = RunResult.STOPPED
                if action.from_estop:
                    self._state.stopped_by_estop = True

        elif isinstance(action, FinishAction):
            if not self._state.run_result:
                self._state.queue_status = QueueStatus.PAUSED
                if action.set_run_status:
                    self._state.run_result = (
                        RunResult.SUCCEEDED
                        if not action.error_details
                        else RunResult.FAILED
                    )
                else:
                    self._state.run_result = RunResult.STOPPED

                if action.error_details:
                    self._state.run_error = self._map_run_exception_to_error_occurrence(
                        action.error_details.error_id,
                        action.error_details.created_at,
                        action.error_details.error,
                    )

        elif isinstance(action, HardwareStoppedAction):
            self._state.queue_status = QueueStatus.PAUSED
            self._state.run_result = self._state.run_result or RunResult.STOPPED
            self._state.run_completed_at = (
                self._state.run_completed_at or action.completed_at
            )

            if action.finish_error_details:
                self._state.finish_error = (
                    self._map_finish_exception_to_error_occurrence(
                        action.finish_error_details.error_id,
                        action.finish_error_details.created_at,
                        action.finish_error_details.error,
                    )
                )

        elif isinstance(action, DoorChangeAction):
            if self._config.block_on_door_open:
                if action.door_state == DoorState.OPEN:
                    self._state.is_door_blocking = True
                    if self._state.queue_status != QueueStatus.SETUP:
                        self._state.queue_status = QueueStatus.PAUSED
                elif action.door_state == DoorState.CLOSED:
                    self._state.is_door_blocking = False

    @staticmethod
    def _map_run_exception_to_error_occurrence(
        error_id: str, created_at: datetime, exception: Exception
    ) -> ErrorOccurrence:
        """Map a fatal exception from the main part of the run to an ErrorOccurrence."""
        if (
            isinstance(exception, ProtocolCommandFailedError)
            and exception.original_error is not None
        ):
            return exception.original_error
        elif isinstance(exception, EnumeratedError):
            return ErrorOccurrence.from_failed(
                id=error_id, createdAt=created_at, error=exception
            )
        else:
            enumerated_wrapper = UnexpectedProtocolError(
                message=str(exception),
                wrapping=[exception],
            )
            return ErrorOccurrence.from_failed(
                id=error_id, createdAt=created_at, error=enumerated_wrapper
            )

    @staticmethod
    def _map_finish_exception_to_error_occurrence(
        error_id: str, created_at: datetime, exception: Exception
    ) -> ErrorOccurrence:
        """Map a fatal exception from the finish phase (drop tip & home) to an ErrorOccurrence."""
        if isinstance(exception, EnumeratedError):
            return ErrorOccurrence.from_failed(
                id=error_id, createdAt=created_at, error=exception
            )
        else:
            enumerated_wrapper = PythonException(exc=exception)
            return ErrorOccurrence.from_failed(
                id=error_id, createdAt=created_at, error=enumerated_wrapper
            )


class CommandView(HasState[CommandState]):
    """Read-only command state view."""

    _state: CommandState

    def __init__(self, state: CommandState) -> None:
        """Initialize the view of command state with its underlying data."""
        self._state = state

    def get(self, command_id: str) -> Command:
        """Get a command by its unique identifier."""
        try:
            return self._state.commands_by_id[command_id].command
        except KeyError:
            raise CommandDoesNotExistError(f"Command {command_id} does not exist")

    def get_all(self) -> List[Command]:
        """Get a list of all commands in state.

        Entries are returned in the order of first-added command to last-added command.
        Replacing a command (to change its status, for example) keeps its place in the
        ordering.
        """
        return [
            self._state.commands_by_id[cid].command
            for cid in self._state.all_command_ids
        ]

    def get_slice(
        self,
        cursor: Optional[int],
        length: int,
    ) -> CommandSlice:
        """Get a subset of commands around a given cursor.

        If the cursor is omitted, a cursor will be selected automatically
        based on the currently running or most recently executed command."
        """
        # TODO(mc, 2022-01-31): this is not the most performant way to implement
        # this; if this becomes a problem, change or the underlying data structure
        # to something that isn't just an OrderedDict
        all_command_ids = self._state.all_command_ids
        commands_by_id = self._state.commands_by_id
        running_command_id = self._state.running_command_id
        queued_command_ids = self._state.queued_command_ids
        total_length = len(all_command_ids)

        if cursor is None:
            if running_command_id is not None:
                cursor = commands_by_id[running_command_id].index
            elif len(queued_command_ids) > 0:
                cursor = commands_by_id[queued_command_ids.head()].index - 1
            elif (
                self._state.run_result
                and self._state.run_result == RunResult.FAILED
                and self._state.failed_command
            ):
                cursor = self._state.failed_command.index
            else:
                cursor = total_length - length

        # start is inclusive, stop is exclusive
        actual_cursor = max(0, min(cursor, total_length - 1))
        stop = min(total_length, actual_cursor + length)
        command_ids = all_command_ids[actual_cursor:stop]
        commands = [commands_by_id[cid].command for cid in command_ids]

        return CommandSlice(
            commands=commands,
            cursor=actual_cursor,
            total_length=total_length,
        )

    def get_error(self) -> Optional[ErrorOccurrence]:
        """Get the run's fatal error, if there was one."""
        run_error = self._state.run_error
        finish_error = self._state.finish_error

        if run_error and finish_error:
            combined_error = ErrorOccurrence.construct(
                id=finish_error.id,
                createdAt=finish_error.createdAt,
                errorType="RunAndFinishFailed",
                detail=(
                    "The run had a fatal error,"
                    " and another error happened while doing post-run cleanup."
                ),
                # TODO(mm, 2023-07-31): Consider adding a low-priority error code so clients can
                # deemphasize this root node, in favor of its children in wrappedErrors.
                errorCode=ErrorCodes.GENERAL_ERROR.value.code,
                wrappedErrors=[
                    run_error,
                    finish_error,
                ],
            )
            return combined_error
        else:
            return run_error or finish_error

    def get_current(self) -> Optional[CurrentCommand]:
        """Return the "current" command, if any.

        The "current" command is the command that is currently executing,
        or the most recent command to have completed.
        """
        if self._state.running_command_id:
            entry = self._state.commands_by_id[self._state.running_command_id]
            return CurrentCommand(
                command_id=entry.command.id,
                command_key=entry.command.key,
                created_at=entry.command.createdAt,
                index=entry.index,
            )

        # TODO(mc, 2022-02-07): this is O(n) in the worst case for no good reason.
        # Resolve prior to JSONv6 support, where this will matter.
        for reverse_index, cid in enumerate(reversed(self._state.all_command_ids)):
            if self.get_command_is_final(cid):
                entry = self._state.commands_by_id[cid]
                return CurrentCommand(
                    command_id=entry.command.id,
                    command_key=entry.command.key,
                    created_at=entry.command.createdAt,
                    index=len(self._state.all_command_ids) - reverse_index - 1,
                )

        return None

    def get_next_to_execute(self) -> Optional[str]:
        """Return the next command in line to be executed.

        Returns:
            The ID of the earliest queued command, if any.

        Raises:
            RunStoppedError: The engine is currently stopped or stopping,
                so it will never run any more commands.
        """
        if self._state.run_result:
            raise RunStoppedError("Engine was stopped")

        # if there is a setup command queued, prioritize it
        next_setup_cmd = self._state.queued_setup_command_ids.head(None)
        if self._state.queue_status != QueueStatus.PAUSED and next_setup_cmd:
            return next_setup_cmd

        # if the queue is running, return the next protocol command
        if self._state.queue_status == QueueStatus.RUNNING:
            return self._state.queued_command_ids.head(None)

        # otherwise we've got nothing to do
        return None

    def get_is_okay_to_clear(self) -> bool:
        """Get whether the engine is stopped or sitting idly so it could be removed."""
        if self.get_is_stopped():
            return True
        elif (
            self.get_status() == EngineStatus.IDLE
            and self._state.running_command_id is None
            and len(self._state.queued_setup_command_ids) == 0
        ):
            return True
        else:
            return False

    def get_is_door_blocking(self) -> bool:
        """Get whether the robot door is open when 'pause on door open' ff is True."""
        return self._state.is_door_blocking

    def get_is_implicitly_active(self) -> bool:
        """Get whether the queue is implicitly active, i.e., never 'played'."""
        return self._state.queue_status == QueueStatus.SETUP

    def get_is_running(self) -> bool:
        """Get whether the protocol is running & queued commands should be executed."""
        return self._state.queue_status == QueueStatus.RUNNING

    def get_command_is_final(self, command_id: str) -> bool:
        """Get whether a given command has reached its final `status`.

        This happens when one of the following is true:

        - Its status is `CommandStatus.SUCCEEDED`.
        - Its status is `CommandStatus.FAILED`.
        - Its status is `CommandStatus.QUEUED` but the run has been requested to stop,
          so the run will never reach it.

        Arguments:
            command_id: Command to check.
        """
        status = self.get(command_id).status

        return (
            status == CommandStatus.SUCCEEDED
            or status == CommandStatus.FAILED
            or (status == CommandStatus.QUEUED and self._state.run_result is not None)
        )

    def get_all_commands_final(self) -> bool:
        """Get whether all commands added so far have reached their final `status`.

        See `get_command_is_final()`.

        Raises:
            CommandExecutionFailedError: if any added command failed, and its `intent` wasn't
            `setup`.
        """
        no_command_running = self._state.running_command_id is None
        no_command_to_execute = (
            self._state.run_result is not None
            or len(self._state.queued_command_ids) == 0
        )

        if no_command_running and no_command_to_execute:
            for command_id in self._state.all_command_ids:
                command = self._state.commands_by_id[command_id].command
                if command.error and command.intent != CommandIntent.SETUP:
                    # TODO(tz, 7-11-23): avoid raising an error and return the status instead
                    raise ProtocolCommandFailedError(
                        original_error=command.error, message=command.error.detail
                    )
            return True
        else:
            return False

    def get_is_stopped(self) -> bool:
        """Get whether an engine stop has completed."""
        return self._state.run_completed_at is not None

    def has_been_played(self) -> bool:
        """Get whether engine has started."""
        return self._state.run_started_at is not None

    def get_is_terminal(self) -> bool:
        """Get whether engine is in a terminal state."""
        return self._state.run_result is not None

    def validate_action_allowed(
        self,
        action: Union[PlayAction, PauseAction, StopAction, QueueCommandAction],
    ) -> Union[PlayAction, PauseAction, StopAction, QueueCommandAction]:
        """Validate whether a given control action is allowed.

        Returns:
            The action, if valid.

        Raises:
            RunStoppedError: The engine has been stopped.
            RobotDoorOpenError: Cannot resume because the front door is open.
            PauseNotAllowedError: The engine is not running, so cannot be paused.
            SetupCommandNotAllowedError: The engine is running, so a setup command
                may not be added.
        """
        if self._state.run_result is not None:
            raise RunStoppedError("The run has already stopped.")

        elif isinstance(action, PlayAction):
            if self.get_status() == EngineStatus.BLOCKED_BY_OPEN_DOOR:
                raise RobotDoorOpenError("Front door or top window is currently open.")

        elif isinstance(action, PauseAction):
            if not self.get_is_running():
                raise PauseNotAllowedError("Cannot pause a run that is not running.")

        elif (
            isinstance(action, QueueCommandAction)
            and action.request.intent == CommandIntent.SETUP
        ):
            if self._state.queue_status != QueueStatus.SETUP:
                raise SetupCommandNotAllowedError(
                    "Setup commands are not allowed after run has started."
                )

        return action

    def get_status(self) -> EngineStatus:
        """Get the current execution status of the engine."""
        if self._state.run_result:
            # The main part of the run is over, or will be over soon.
            # Have we also completed the post-run finish steps (homing and dropping tips)?
            if self.get_is_stopped():
                # Post-run finish steps have completed. Calculate the engine's final status,
                # taking into account any failures in the run or the post-run finish steps.
                if (
                    self._state.run_result == RunResult.FAILED
                    or self._state.finish_error is not None
                ):
                    return EngineStatus.FAILED
                elif self._state.run_result == RunResult.SUCCEEDED:
                    return EngineStatus.SUCCEEDED
                else:
                    return EngineStatus.STOPPED
            else:
                # Post-run finish steps have not yet completed,
                # and we may even still be executing commands.
                return (
                    EngineStatus.STOP_REQUESTED
                    if self._state.run_result == RunResult.STOPPED
                    else EngineStatus.FINISHING
                )

        elif self._state.queue_status == QueueStatus.RUNNING:
            return EngineStatus.RUNNING

        elif self._state.queue_status == QueueStatus.PAUSED:
            if self._state.is_door_blocking:
                return EngineStatus.BLOCKED_BY_OPEN_DOOR
            else:
                return EngineStatus.PAUSED

        return EngineStatus.IDLE

    def get_latest_command_hash(self) -> Optional[str]:
        """Get the command hash of the last queued command, if any."""
        return self._state.latest_command_hash



"""Convert deck slots into the preferred style for a robot.

The deck slots on an OT-2 are labeled like "1", "2", ..., and on an OT-3 they're labeled like
"D1," "D2", ....

When Protocol Engine takes a deck slot as input, we generally want it to accept either style
of label. This helps make protocols more portable across robot types.

But, Protocol Engine should then immediately convert it to the "correct" style for the robot that
it's controlling or simulating. This makes it simpler to consume the robot's HTTP API,
and it makes it easier for us to reason about Protocol Engine's internal state.

This module does that conversion, for any Protocol Engine input that contains a reference to a
deck slot.
"""


from typing import Any, Callable, Dict, Type

from opentrons_shared_data.robot.dev_types import RobotType

from . import commands
from .types import (
    OFF_DECK_LOCATION,
    DeckSlotLocation,
    LabwareLocation,
    AddressableAreaLocation,
    LabwareOffsetCreate,
    ModuleLocation,
    OnLabwareLocation,
)


def standardize_labware_offset(
    original: LabwareOffsetCreate, robot_type: RobotType
) -> LabwareOffsetCreate:
    """Convert the deck slot in the given `LabwareOffsetCreate` to match the given robot type."""
    return original.copy(
        update={
            "location": original.location.copy(
                update={
                    "slotName": original.location.slotName.to_equivalent_for_robot_type(
                        robot_type
                    )
                }
            )
        }
    )


def standardize_command(
    original: commands.CommandCreate, robot_type: RobotType
) -> commands.CommandCreate:
    """Convert any deck slots in the given `CommandCreate` to match the given robot type."""
    try:
        standardize = _standardize_command_functions[type(original)]
    except KeyError:
        return original
    else:
        return standardize(original, robot_type)


# Command-specific standardization:
#
# Our use of .copy(update=...) in these implementations, instead of .construct(...), is a tradeoff.
# .construct() would give us better type-checking for the fields that we set,
# but .copy(update=...) avoids the hazard of forgetting to set fields that have defaults.


def _standardize_load_labware(
    original: commands.LoadLabwareCreate, robot_type: RobotType
) -> commands.LoadLabwareCreate:
    params = original.params.copy(
        update={
            "location": _standardize_labware_location(
                original.params.location, robot_type
            )
        }
    )
    return original.copy(update={"params": params})


def _standardize_load_module(
    original: commands.LoadModuleCreate, robot_type: RobotType
) -> commands.LoadModuleCreate:
    params = original.params.copy(
        update={
            "location": _standardize_deck_slot_location(
                original.params.location, robot_type
            )
        }
    )
    return original.copy(update={"params": params})


def _standardize_move_labware(
    original: commands.MoveLabwareCreate, robot_type: RobotType
) -> commands.MoveLabwareCreate:
    params = original.params.copy(
        update={
            "newLocation": _standardize_labware_location(
                original.params.newLocation, robot_type
            )
        }
    )
    return original.copy(update={"params": params})


_standardize_command_functions: Dict[
    Type[commands.CommandCreate], Callable[[Any, RobotType], commands.CommandCreate]
] = {
    commands.LoadLabwareCreate: _standardize_load_labware,
    commands.LoadModuleCreate: _standardize_load_module,
    commands.MoveLabwareCreate: _standardize_move_labware,
}


# Helpers:


def _standardize_labware_location(
    original: LabwareLocation, robot_type: RobotType
) -> LabwareLocation:
    if isinstance(original, DeckSlotLocation):
        return _standardize_deck_slot_location(original, robot_type)
    elif (
        isinstance(
            original, (ModuleLocation, OnLabwareLocation, AddressableAreaLocation)
        )
        or original == OFF_DECK_LOCATION
    ):
        return original


def _standardize_deck_slot_location(
    original: DeckSlotLocation, robot_type: RobotType
) -> DeckSlotLocation:
    return original.copy(
        update={"slotName": original.slotName.to_equivalent_for_robot_type(robot_type)}
    )



import typing
from typing_extensions import Literal, Final
from dataclasses import dataclass, field
from datetime import datetime

from opentrons.config import feature_flags as ff
from opentrons.config.robot_configs import (
    default_pipette_offset,
    default_gripper_calibration_offset,
)
from opentrons.types import Point, Mount

# TODO change this when imports are fixed
from opentrons.calibration_storage.ot3.pipette_offset import (
    save_pipette_calibration,
    get_pipette_offset,
)
from opentrons.calibration_storage import (
    types as cal_top_types,
    gripper_offset,
)
from opentrons.hardware_control.types import OT3Mount

PIPETTE_OFFSET_CONSISTENCY_LIMIT: Final = 1.5

# These type aliases aid typechecking in tests that work the same on this and
# the hardware_control.instruments.ot2 variant
SourceType = cal_top_types.SourceType
CalibrationStatus = cal_top_types.CalibrationStatus


@dataclass
class InconsistentPipetteOffsets:
    kind: Literal["inconsistent-pipette-offset"]
    offsets: typing.Dict[Mount, Point]
    limit: float


ReasonabilityCheckFailure = typing.Union[InconsistentPipetteOffsets]


@dataclass
class PipetteOffsetByPipetteMount:
    """
    Class to store pipette offset without pipette and mount info
    """

    offset: Point
    source: SourceType
    status: CalibrationStatus
    last_modified: typing.Optional[datetime] = None


@dataclass
class PipetteOffsetSummary(PipetteOffsetByPipetteMount):
    reasonability_check_failures: typing.List[ReasonabilityCheckFailure] = field(
        default_factory=list
    )


@dataclass
class GripperCalibrationOffset:
    """
    Class to store gripper offset calibration with gripper info

    TODO(lc 09/15/2022) this can probably be combined with
    pipette offset mount.
    """

    offset: Point
    source: SourceType
    status: CalibrationStatus
    last_modified: typing.Optional[datetime] = None


def load_pipette_offset(
    pip_id: typing.Optional[str], mount: OT3Mount
) -> PipetteOffsetByPipetteMount:
    # load default if pipette offset data do not exist
    pip_cal_obj = PipetteOffsetByPipetteMount(
        offset=Point(*default_pipette_offset()),
        source=SourceType.default,
        status=CalibrationStatus(),
    )
    # TODO this can be removed once we switch to using
    # ot3 pipette types in the ot3 hardware controller.
    if isinstance(mount, OT3Mount):
        checked_mount = mount.to_mount()
    else:
        checked_mount = mount
    if pip_id:
        pip_offset_data = get_pipette_offset(pip_id, checked_mount)
        if pip_offset_data:
            return PipetteOffsetByPipetteMount(
                offset=pip_offset_data.offset,
                last_modified=pip_offset_data.lastModified,
                source=pip_offset_data.source,
                status=CalibrationStatus(
                    markedAt=pip_offset_data.status.markedAt,
                    markedBad=pip_offset_data.status.markedBad,
                    source=pip_offset_data.status.source,
                ),
            )
    return pip_cal_obj


def save_pipette_offset_calibration(
    pip_id: typing.Optional[str], mount: typing.Union[Mount, OT3Mount], offset: Point
) -> None:
    # TODO this can be removed once we switch to using
    # ot3 pipette types in the ot3 hardware controller.
    if isinstance(mount, OT3Mount):
        checked_mount = mount.to_mount()
    else:
        checked_mount = mount
    if pip_id:
        save_pipette_calibration(offset, pip_id, checked_mount)


def load_gripper_calibration_offset(
    gripper_id: typing.Optional[str],
) -> GripperCalibrationOffset:
    # load default if gripper offset data do not exist
    grip_cal_obj = GripperCalibrationOffset(
        offset=Point(*default_gripper_calibration_offset()),
        source=SourceType.default,
        status=CalibrationStatus(),
    )
    if gripper_id and ff.enable_ot3_hardware_controller():
        grip_offset_data = gripper_offset.get_gripper_calibration_offset(gripper_id)
        if grip_offset_data:
            return GripperCalibrationOffset(
                offset=grip_offset_data.offset,
                last_modified=grip_offset_data.lastModified,
                source=grip_offset_data.source,
                status=cal_top_types.CalibrationStatus(
                    markedAt=grip_offset_data.status.markedAt,
                    markedBad=grip_offset_data.status.markedBad,
                    source=grip_offset_data.status.source,
                ),
            )
    return grip_cal_obj


def save_gripper_calibration_offset(
    gripper_id: typing.Optional[str], delta: Point
) -> None:
    if gripper_id and ff.enable_ot3_hardware_controller():
        gripper_offset.save_gripper_calibration(delta, gripper_id)


def check_instrument_offset_reasonability(
    left_offset: Point, right_offset: Point
) -> typing.List[ReasonabilityCheckFailure]:
    if (
        not left_offset
        or left_offset == Point(0, 0, 0)
        or not right_offset
        or right_offset == Point(0, 0, 0)
    ):
        return []
    diff = left_offset - right_offset
    if any(abs(d) > PIPETTE_OFFSET_CONSISTENCY_LIMIT for d in diff):
        return [
            InconsistentPipetteOffsets(
                "inconsistent-pipette-offset",
                {Mount.LEFT: left_offset, Mount.RIGHT: right_offset},
                PIPETTE_OFFSET_CONSISTENCY_LIMIT,
            )
        ]
    return []



"""Public protocol run data models."""
from pydantic import BaseModel, Field
from typing import List, Optional
from datetime import datetime

from ..errors import ErrorOccurrence
from ..types import (
    EngineStatus,
    LoadedLabware,
    LabwareOffset,
    LoadedModule,
    LoadedPipette,
    Liquid,
)


class StateSummary(BaseModel):
    """Data from a protocol run."""

    status: EngineStatus
    # errors is a list for historical reasons. (This model needs to stay compatible with
    # robot-server's database.) It shouldn't have more than 1 element.
    errors: List[ErrorOccurrence]
    labware: List[LoadedLabware]
    pipettes: List[LoadedPipette]
    modules: List[LoadedModule]
    labwareOffsets: List[LabwareOffset]
    startedAt: Optional[datetime]
    completedAt: Optional[datetime]
    liquids: List[Liquid] = Field(default_factory=list)



"""Tip pickup and drop procedures."""
from typing import Optional, Dict
from typing_extensions import Protocol as TypingProtocol

from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.types import FailedTipStateCheck
from opentrons_shared_data.errors.exceptions import (
    CommandPreconditionViolated,
    CommandParameterLimitViolated,
    PythonException,
)

from ..resources import LabwareDataProvider, ensure_ot3_hardware
from ..state import StateView
from ..types import TipGeometry, TipPresenceStatus
from ..errors import (
    HardwareNotSupportedError,
    TipNotAttachedError,
    TipAttachedError,
    ProtocolEngineError,
)


PRIMARY_NOZZLE_TO_ENDING_NOZZLE_MAP = {
    "A1": {"COLUMN": "H1", "ROW": "A12"},
    "H1": {"COLUMN": "A1", "ROW": "H12"},
    "A12": {"COLUMN": "H12", "ROW": "A1"},
    "H12": {"COLUMN": "A12", "ROW": "H1"},
}


class TipHandler(TypingProtocol):
    """Pick up and drop tips."""

    async def available_for_nozzle_layout(
        self,
        pipette_id: str,
        style: str,
        primary_nozzle: Optional[str] = None,
        front_right_nozzle: Optional[str] = None,
    ) -> Dict[str, str]:
        """Check nozzle layout is compatible with the pipette.

        Returns:
            A dict of nozzles used to configure the pipette.
        """
        ...

    async def pick_up_tip(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
    ) -> TipGeometry:
        """Pick up the named tip.

        Pipette should be in place over the named tip prior to calling this method.

        Returns:
            Tip geometry of the picked up tip.
        """
        ...

    async def drop_tip(self, pipette_id: str, home_after: Optional[bool]) -> None:
        """Drop the attached tip into the named location.

        Pipette should be in place over the destination prior to calling this method.
        """

    async def add_tip(self, pipette_id: str, tip: TipGeometry) -> None:
        """Tell the Hardware API that a tip is attached."""

    async def get_tip_presence(self, pipette_id: str) -> TipPresenceStatus:
        """Get tip presence status on the pipette."""

    async def verify_tip_presence(
        self, pipette_id: str, expected: TipPresenceStatus
    ) -> None:
        """Verify the expected tip presence status."""


async def _available_for_nozzle_layout(
    channels: int,
    style: str,
    primary_nozzle: Optional[str],
    front_right_nozzle: Optional[str],
) -> Dict[str, str]:
    """Check nozzle layout is compatible with the pipette.

    Returns:
        A dict of nozzles used to configure the pipette.
    """
    if channels == 1:
        raise CommandPreconditionViolated(
            message=f"Cannot configure nozzle layout with a {channels} channel pipette."
        )
    if style == "ALL":
        return {}
    if style == "ROW" and channels == 8:
        raise CommandParameterLimitViolated(
            command_name="configure_nozzle_layout",
            parameter_name="RowNozzleLayout",
            limit_statement="RowNozzleLayout is incompatible with {channels} channel pipettes.",
            actual_value=str(primary_nozzle),
        )
    if not primary_nozzle:
        return {"primary_nozzle": "A1"}
    if style == "SINGLE":
        return {"primary_nozzle": primary_nozzle}
    if not front_right_nozzle:
        return {
            "primary_nozzle": primary_nozzle,
            "front_right_nozzle": PRIMARY_NOZZLE_TO_ENDING_NOZZLE_MAP[primary_nozzle][
                style
            ],
        }
    return {
        "primary_nozzle": primary_nozzle,
        "front_right_nozzle": front_right_nozzle,
    }


class HardwareTipHandler(TipHandler):
    """Pick up and drop tips, using the Hardware API."""

    def __init__(
        self,
        state_view: StateView,
        hardware_api: HardwareControlAPI,
        labware_data_provider: Optional[LabwareDataProvider] = None,
    ) -> None:
        self._hardware_api = hardware_api
        self._labware_data_provider = labware_data_provider or LabwareDataProvider()
        self._state_view = state_view

    async def available_for_nozzle_layout(
        self,
        pipette_id: str,
        style: str,
        primary_nozzle: Optional[str] = None,
        front_right_nozzle: Optional[str] = None,
    ) -> Dict[str, str]:
        """Returns configuration for nozzle layout to pass to configure_nozzle_layout."""
        if self._state_view.pipettes.get_attached_tip(pipette_id):
            raise CommandPreconditionViolated(
                message=f"Cannot configure nozzle layout of {str(self)} while it has tips attached."
            )
        channels = self._state_view.pipettes.get_channels(pipette_id)
        return await _available_for_nozzle_layout(
            channels, style, primary_nozzle, front_right_nozzle
        )

    async def pick_up_tip(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
    ) -> TipGeometry:
        """Pick up a tip at the current location using the Hardware API."""
        hw_mount = self._state_view.pipettes.get_mount(pipette_id).to_hw_mount()

        nominal_tip_geometry = self._state_view.geometry.get_nominal_tip_geometry(
            pipette_id=pipette_id, labware_id=labware_id, well_name=well_name
        )

        actual_tip_length = await self._labware_data_provider.get_calibrated_tip_length(
            pipette_serial=self._state_view.pipettes.get_serial_number(pipette_id),
            labware_definition=self._state_view.labware.get_definition(labware_id),
            nominal_fallback=nominal_tip_geometry.length,
        )

        await self._hardware_api.pick_up_tip(
            mount=hw_mount,
            tip_length=actual_tip_length,
            presses=None,
            increment=None,
        )
        await self.verify_tip_presence(pipette_id, TipPresenceStatus.PRESENT)

        self._hardware_api.set_current_tiprack_diameter(
            mount=hw_mount,
            tiprack_diameter=nominal_tip_geometry.diameter,
        )

        self._hardware_api.set_working_volume(
            mount=hw_mount,
            tip_volume=nominal_tip_geometry.volume,
        )

        return TipGeometry(
            length=actual_tip_length,
            diameter=nominal_tip_geometry.diameter,
            volume=nominal_tip_geometry.volume,
        )

    async def drop_tip(self, pipette_id: str, home_after: Optional[bool]) -> None:
        """Drop a tip at the current location using the Hardware API."""
        hw_mount = self._state_view.pipettes.get_mount(pipette_id).to_hw_mount()

        # Let the hardware controller handle defaulting home_after since its behavior
        # differs between machines
        if home_after is not None:
            kwargs = {"home_after": home_after}
        else:
            kwargs = {}

        await self._hardware_api.drop_tip(mount=hw_mount, **kwargs)
        await self.verify_tip_presence(pipette_id, TipPresenceStatus.ABSENT)

    async def add_tip(self, pipette_id: str, tip: TipGeometry) -> None:
        """Tell the Hardware API that a tip is attached."""
        hw_mount = self._state_view.pipettes.get_mount(pipette_id).to_hw_mount()

        await self._hardware_api.add_tip(mount=hw_mount, tip_length=tip.length)

        self._hardware_api.set_current_tiprack_diameter(
            mount=hw_mount,
            tiprack_diameter=tip.diameter,
        )

        self._hardware_api.set_working_volume(
            mount=hw_mount,
            tip_volume=tip.volume,
        )

    async def get_tip_presence(self, pipette_id: str) -> TipPresenceStatus:
        """Get the tip presence status of the pipette."""
        try:
            ot3api = ensure_ot3_hardware(hardware_api=self._hardware_api)

            hw_mount = self._state_view.pipettes.get_mount(pipette_id).to_hw_mount()

            status = await ot3api.get_tip_presence_status(hw_mount)
            return TipPresenceStatus.from_hw_state(status)
        except HardwareNotSupportedError:
            # Tip presence sensing is not supported on the OT2
            return TipPresenceStatus.UNKNOWN

    async def verify_tip_presence(
        self, pipette_id: str, expected: TipPresenceStatus
    ) -> None:
        """Verify the expecterd tip presence status of the pipette.

        This function will raise an exception if the specified tip presence status
        isn't matched.
        """
        try:
            ot3api = ensure_ot3_hardware(hardware_api=self._hardware_api)
            hw_mount = self._state_view.pipettes.get_mount(pipette_id).to_hw_mount()
            await ot3api.verify_tip_presence(hw_mount, expected.to_hw_state())
        except HardwareNotSupportedError:
            # Tip presence sensing is not supported on the OT2
            pass
        except FailedTipStateCheck as e:
            if expected == TipPresenceStatus.ABSENT:
                raise TipAttachedError(wrapping=[PythonException(e)])
            elif expected == TipPresenceStatus.PRESENT:
                raise TipNotAttachedError(wrapping=[PythonException(e)])
            else:
                raise ProtocolEngineError(
                    message="Unknown tip status in tip status check",
                    wrapping=[PythonException(e)],
                )


class VirtualTipHandler(TipHandler):
    """Pick up and drop tips, using a virtual pipette."""

    def __init__(self, state_view: StateView) -> None:
        self._state_view = state_view

    async def pick_up_tip(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
    ) -> TipGeometry:
        """Pick up a tip at the current location using a virtual pipette.

        - Fetch nominal tip geometry
        - Check that there's no tip currently attached
        """
        nominal_tip_geometry = self._state_view.geometry.get_nominal_tip_geometry(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
        )

        self._state_view.pipettes.validate_tip_state(
            pipette_id=pipette_id,
            expected_has_tip=False,
        )

        return nominal_tip_geometry

    async def available_for_nozzle_layout(
        self,
        pipette_id: str,
        style: str,
        primary_nozzle: Optional[str] = None,
        front_right_nozzle: Optional[str] = None,
    ) -> Dict[str, str]:
        """Returns configuration for nozzle layout to pass to configure_nozzle_layout."""
        if self._state_view.pipettes.get_attached_tip(pipette_id):
            raise CommandPreconditionViolated(
                message=f"Cannot configure nozzle layout of {str(self)} while it has tips attached."
            )
        channels = self._state_view.pipettes.get_channels(pipette_id)
        return await _available_for_nozzle_layout(
            channels, style, primary_nozzle, front_right_nozzle
        )

    async def drop_tip(
        self,
        pipette_id: str,
        home_after: Optional[bool],
    ) -> None:
        """Pick up a tip at the current location using a virtual pipette.

        - Check that there's no tip currently attached
        """
        self._state_view.pipettes.validate_tip_state(
            pipette_id=pipette_id,
            expected_has_tip=True,
        )

    async def add_tip(self, pipette_id: str, tip: TipGeometry) -> None:
        """Add a tip using a virtual pipette.

        This should not be called when using virtual pipettes.
        """
        assert False, "TipHandler.add_tip should not be used with virtual pipettes"

    async def verify_tip_presence(
        self, pipette_id: str, expected: TipPresenceStatus
    ) -> None:
        """Verify tip presence.

        This should not be called when using virtual pipettes.
        """

    async def get_tip_presence(self, pipette_id: str) -> TipPresenceStatus:
        """Get tip presence.

        This is a check to the physical machine's sensors  and should not be
        called on a virtual pipette.
        """
        raise RuntimeError("Do not call VirtualTipHandler.get_tip_presence")


def create_tip_handler(
    state_view: StateView, hardware_api: HardwareControlAPI
) -> TipHandler:
    """Create a tip handler."""
    return (
        HardwareTipHandler(state_view=state_view, hardware_api=hardware_api)
        if state_view.config.use_virtual_pipettes is False
        else VirtualTipHandler(state_view=state_view)
    )



from __future__ import annotations

import functools

""" Classes and functions for pipette state tracking
"""
import logging
from typing import Any, Dict, Optional, Set, Tuple, Union, cast

from opentrons_shared_data.pipette.pipette_definition import (
    PipetteConfigurations,
    PlungerPositions,
    MotorConfigurations,
    SupportedTipsDefinition,
    PickUpTipConfigurations,
    DropTipConfigurations,
    PipetteModelVersionType,
    PipetteNameType,
    PipetteLiquidPropertiesDefinition,
)
from opentrons_shared_data.pipette import (
    load_data as load_pipette_data,
    types as pip_types,
)
from opentrons_shared_data.errors.exceptions import (
    InvalidLiquidClassName,
    CommandPreconditionViolated,
)


from opentrons.types import Point, Mount
from opentrons.config import robot_configs
from opentrons.config.types import RobotConfig
from opentrons.drivers.types import MoveSplit
from ..instrument_abc import AbstractInstrument
from ..instrument_helpers import (
    piecewise_volume_conversion,
    PIPETTING_FUNCTION_FALLBACK_VERSION,
    PIPETTING_FUNCTION_LATEST_VERSION,
)
from .instrument_calibration import (
    PipetteOffsetByPipetteMount,
    load_pipette_offset,
)
from opentrons.hardware_control.types import (
    CriticalPoint,
    BoardRevision,
)
from opentrons.hardware_control.errors import InvalidCriticalPoint
from opentrons.hardware_control import nozzle_manager


from opentrons_shared_data.pipette.dev_types import (
    UlPerMmAction,
    PipetteName,
    PipetteModel,
)
from opentrons.hardware_control.dev_types import InstrumentHardwareConfigs


RECONFIG_KEYS = {"quirks"}


mod_log = logging.getLogger(__name__)

# TODO (lc 11-1-2022) Once we unify calibration loading
# for the hardware controller, we will be able to
# unify the pipette classes again.


class Pipette(AbstractInstrument[PipetteConfigurations]):
    """A class to gather and track pipette state and configs.

    This class should not touch hardware or call back out to the hardware
    control API. Its only purpose is to gather state.
    """

    DictType = Dict[
        str, Union[str, float, bool]
    ]  # spp: as_dict() has value items that aren't Union[str, float, bool]..
    #: The type of this data class as a dict

    def __init__(
        self,
        config: PipetteConfigurations,
        pipette_offset_cal: PipetteOffsetByPipetteMount,
        pipette_id: Optional[str] = None,
        use_old_aspiration_functions: bool = False,
    ) -> None:
        self._config = config
        self._config_as_dict = config.dict()
        self._pipette_offset = pipette_offset_cal
        self._pipette_type = self._config.pipette_type
        self._pipette_version = self._config.version
        self._max_channels = self._config.channels
        self._backlash_distance = config.backlash_distance
        self._pick_up_configurations = config.pick_up_tip_configurations

        self._liquid_class_name = pip_types.LiquidClasses.default
        self._liquid_class = self._config.liquid_properties[self._liquid_class_name]

        # TODO (lc 12-05-2022) figure out how we can safely deprecate "name" and "model"
        self._pipette_name = PipetteNameType(
            pipette_type=config.pipette_type,
            pipette_channels=config.channels,
            pipette_generation=config.display_category,
        )
        self._acting_as = self._pipette_name
        self._pipette_model = PipetteModelVersionType(
            pipette_type=config.pipette_type,
            pipette_channels=config.channels,
            pipette_version=config.version,
        )
        self._nozzle_offset = self._config.nozzle_offset
        self._nozzle_manager = (
            nozzle_manager.NozzleConfigurationManager.build_from_config(self._config)
        )
        self._current_volume = 0.0
        self._working_volume = float(self._liquid_class.max_volume)
        self._current_tip_length = 0.0
        self._current_tiprack_diameter = 0.0
        self._has_tip = False
        self._pipette_id = pipette_id
        self._log = mod_log.getChild(
            self._pipette_id if self._pipette_id else "<unknown>"
        )
        self._log.info(
            "loaded: {}, pipette offset: {}".format(
                self._pipette_model, self._pipette_offset.offset
            )
        )
        self.ready_to_aspirate = False
        #: True if ready to aspirate

        self._active_tip_settings = self._liquid_class.supported_tips[
            pip_types.PipetteTipType(self._liquid_class.max_volume)
        ]
        self._fallback_tip_length = self._active_tip_settings.default_tip_length

        self._aspirate_flow_rate = (
            self._active_tip_settings.default_aspirate_flowrate.default
        )
        self._dispense_flow_rate = (
            self._active_tip_settings.default_dispense_flowrate.default
        )
        self._blow_out_flow_rate = (
            self._active_tip_settings.default_blowout_flowrate.default
        )

        self._tip_overlap_lookup = self._liquid_class.tip_overlap_dictionary

        if use_old_aspiration_functions:
            self._pipetting_function_version = PIPETTING_FUNCTION_FALLBACK_VERSION
        else:
            self._pipetting_function_version = PIPETTING_FUNCTION_LATEST_VERSION

    @property
    def push_out_volume(self) -> float:
        return self._active_tip_settings.default_push_out_volume

    def act_as(self, name: PipetteNameType) -> None:
        """Reconfigure to act as ``name``. ``name`` must be either the
        actual name of the pipette, or a name in its back-compatibility
        config.
        """
        if name == self._acting_as:
            return

        str_name = f"{name.pipette_type.name}_{str(name.pipette_channels)}"
        assert str_name in self._config.pipette_backcompat_names + [
            self.name
        ], f"{self.name} is not back-compatible with {name}"

        liquid_model = load_pipette_data.load_liquid_model(
            name.pipette_type, name.pipette_channels, name.get_version()
        )
        # TODO need to grab name config here to deal with act as test
        self._liquid_class.max_volume = liquid_model["default"].max_volume
        self._liquid_class.min_volume = liquid_model["default"].min_volume
        self.working_volume = liquid_model["default"].max_volume
        self.update_config_item(
            {
                "min_volume": liquid_model["default"].min_volume,
                "max_volume": liquid_model["default"].max_volume,
            },
            pip_types.LiquidClasses.default,
        )

    @property
    def acting_as(self) -> PipetteNameType:
        return self._acting_as

    @property
    def config(self) -> PipetteConfigurations:
        return self._config

    @property
    def liquid_class(self) -> PipetteLiquidPropertiesDefinition:
        return self._liquid_class

    @property
    def liquid_class_name(self) -> pip_types.LiquidClasses:
        return self._liquid_class_name

    @property
    def nozzle_offset(self) -> Point:
        return self._nozzle_manager.starting_nozzle_offset

    @property
    def nozzle_manager(self) -> nozzle_manager.NozzleConfigurationManager:
        return self._nozzle_manager

    @property
    def pipette_offset(self) -> PipetteOffsetByPipetteMount:
        return self._pipette_offset

    @property
    def tip_overlap(self) -> Dict[str, float]:
        return self._tip_overlap_lookup

    @property
    def channels(self) -> pip_types.PipetteChannelType:
        return self._max_channels

    @property
    def plunger_positions(self) -> PlungerPositions:
        return self._config.plunger_positions_configurations[self._liquid_class_name]

    @property
    def plunger_motor_current(self) -> MotorConfigurations:
        return self._config.plunger_motor_configurations

    @property
    def pick_up_configurations(self) -> PickUpTipConfigurations:
        return self._config.pick_up_tip_configurations

    @pick_up_configurations.setter
    def pick_up_configurations(self, pick_up_configs: PickUpTipConfigurations) -> None:
        self._pick_up_configurations = pick_up_configs

    @property
    def drop_configurations(self) -> DropTipConfigurations:
        return self._config.drop_tip_configurations

    @property
    def active_tip_settings(self) -> SupportedTipsDefinition:
        return self._active_tip_settings

    def update_config_item(
        self,
        elements: Dict[str, Any],
        liquid_class: Optional[pip_types.LiquidClasses] = None,
    ) -> None:
        self._log.info(f"updated config: {elements}")
        self._config = load_pipette_data.update_pipette_configuration(
            self._config, elements, liquid_class
        )
        # Update the cached dict representation
        self._config_as_dict = self._config.dict()

    def reload_configurations(self) -> None:
        self._config = load_pipette_data.load_definition(
            self._pipette_model.pipette_type,
            self._pipette_model.pipette_channels,
            self._pipette_model.pipette_version,
        )
        self._config_as_dict = self._config.dict()

    def reset_state(self) -> None:
        self._current_volume = 0.0
        self._working_volume = float(self.liquid_class.max_volume)
        self._current_tip_length = 0.0
        self._current_tiprack_diameter = 0.0
        self._has_tip = False
        self.ready_to_aspirate = False
        #: True if ready to aspirate
        self._active_tip_settings = self.liquid_class.supported_tips[
            pip_types.PipetteTipType(self._working_volume)
        ]
        self._fallback_tip_length = self.active_tip_settings.default_tip_length

        self._aspirate_flow_rate = (
            self.active_tip_settings.default_aspirate_flowrate.default
        )
        self._dispense_flow_rate = (
            self.active_tip_settings.default_dispense_flowrate.default
        )
        self._blow_out_flow_rate = (
            self.active_tip_settings.default_blowout_flowrate.default
        )

        self._tip_overlap_lookup = self.liquid_class.tip_overlap_dictionary
        self._nozzle_manager = (
            nozzle_manager.NozzleConfigurationManager.build_from_config(self._config)
        )

    def reset_pipette_offset(self, mount: Mount, to_default: bool) -> None:
        """Reset the pipette offset to system defaults."""
        if to_default:
            self._pipette_offset = load_pipette_offset(pip_id=None, mount=mount)
        else:
            self._pipette_offset = load_pipette_offset(self._pipette_id, mount)

    def save_pipette_offset(self, mount: Mount, offset: Point) -> None:
        """Update the pipette offset to a new value."""
        # TODO (lc 10-31-2022) We should have this command be supported properly by
        # ot-3 and ot-2 when we split out the pipette class
        self._pipette_offset = load_pipette_offset(self._pipette_id, mount)

    @property
    def name(self) -> PipetteName:
        return cast(PipetteName, f"{self._pipette_name}")

    @property
    def model(self) -> PipetteModel:
        return cast(PipetteModel, f"{self._pipette_model}")

    @property
    def pipette_type(self) -> pip_types.PipetteModelType:
        return self._pipette_type

    @property
    def pipette_id(self) -> Optional[str]:
        return self._pipette_id

    def critical_point(self, cp_override: Optional[CriticalPoint] = None) -> Point:
        """
        The vector from the pipette's origin to its critical point. The
        critical point for a pipette is the end of the nozzle if no tip is
        attached, or the end of the tip if a tip is attached.

        If `cp_override` is specified and valid - so is either
        :py:attr:`CriticalPoint.NOZZLE` or :py:attr:`CriticalPoint.TIP` when
        we have a tip, or :py:attr:`CriticalPoint.XY_CENTER` - the specified
        critical point will be used.
        """

        if cp_override in [
            CriticalPoint.GRIPPER_JAW_CENTER,
            CriticalPoint.GRIPPER_FRONT_CALIBRATION_PIN,
            CriticalPoint.GRIPPER_REAR_CALIBRATION_PIN,
        ]:
            raise InvalidCriticalPoint(cp_override.name, "pipette")

        instr = Point(*self._pipette_offset.offset)
        cp_with_tip_length = self._nozzle_manager.critical_point_with_tip_length(
            cp_override,
            self.current_tip_length if cp_override != CriticalPoint.NOZZLE else 0.0,
        )

        cp = cp_with_tip_length + instr

        if self._log.isEnabledFor(logging.DEBUG):
            info_str = "cp: {}{}: {} (from: ".format(
                cp_override, " (from override)" if cp_override else "", cp
            )
            info_str += "model offset: {} + instrument offset: {}".format(
                cp_with_tip_length, instr
            )
            info_str += " - tip_length: {}".format(self.current_tip_length)
            info_str += ")"
            self._log.debug(info_str)

        return cp

    @property
    def current_volume(self) -> float:
        """The amount of liquid currently aspirated"""
        return self._current_volume

    @property
    def current_tip_length(self) -> float:
        """The length of the current tip attached (0.0 if no tip)"""
        return self._current_tip_length

    @current_tip_length.setter
    def current_tip_length(self, tip_length: float) -> None:
        self._current_tip_length = tip_length

    @property
    def current_tiprack_diameter(self) -> float:
        """The diameter of the current tip rack (0.0 if no tip)"""
        return self._current_tiprack_diameter

    @current_tiprack_diameter.setter
    def current_tiprack_diameter(self, diameter: float) -> None:
        self._current_tiprack_diameter = diameter

    @property
    def aspirate_flow_rate(self) -> float:
        """Current active flow rate (not config value)"""
        return self._aspirate_flow_rate

    @aspirate_flow_rate.setter
    def aspirate_flow_rate(self, new_flow_rate: float) -> None:
        assert new_flow_rate > 0
        self._aspirate_flow_rate = new_flow_rate

    @property
    def dispense_flow_rate(self) -> float:
        """Current active flow rate (not config value)"""
        return self._dispense_flow_rate

    @dispense_flow_rate.setter
    def dispense_flow_rate(self, new_flow_rate: float) -> None:
        assert new_flow_rate > 0
        self._dispense_flow_rate = new_flow_rate

    @property
    def blow_out_flow_rate(self) -> float:
        """Current active flow rate (not config value)"""
        return self._blow_out_flow_rate

    @blow_out_flow_rate.setter
    def blow_out_flow_rate(self, new_flow_rate: float) -> None:
        assert new_flow_rate > 0
        self._blow_out_flow_rate = new_flow_rate

    @property
    def aspirate_flow_rates_lookup(self) -> Dict[str, float]:
        return self._active_tip_settings.default_aspirate_flowrate.values_by_api_level

    @property
    def dispense_flow_rates_lookup(self) -> Dict[str, float]:
        return self._active_tip_settings.default_dispense_flowrate.values_by_api_level

    @property
    def blow_out_flow_rates_lookup(self) -> Dict[str, float]:
        return self.active_tip_settings.default_blowout_flowrate.values_by_api_level

    @property
    def working_volume(self) -> float:
        """The working volume of the pipette"""
        return self._working_volume

    @working_volume.setter
    def working_volume(self, tip_volume: float) -> None:
        """The working volume is the current tip max volume"""
        self._working_volume = min(self.liquid_class.max_volume, tip_volume)
        tip_size_type = pip_types.PipetteTipType.check_and_return_type(
            int(self._working_volume), self.liquid_class.max_volume
        )
        self._active_tip_settings = self.liquid_class.supported_tips[tip_size_type]
        self._fallback_tip_length = self._active_tip_settings.default_tip_length

    @property
    def minimum_volume(self) -> float:
        """The smallest controllable volume the pipette can handle with the current liquid class.."""
        return self.liquid_class.min_volume

    @property
    def available_volume(self) -> float:
        """The amount of liquid possible to aspirate"""
        return self.working_volume - self.current_volume

    def set_current_volume(self, new_volume: float) -> None:
        assert new_volume >= 0
        assert new_volume <= self.working_volume
        self._current_volume = new_volume

    def add_current_volume(self, volume_incr: float) -> None:
        assert self.ok_to_add_volume(volume_incr)
        self._current_volume += volume_incr

    def remove_current_volume(self, volume_incr: float) -> None:
        assert self._current_volume >= volume_incr
        self._current_volume -= volume_incr

    def ok_to_add_volume(self, volume_incr: float) -> bool:
        return self.current_volume + volume_incr <= self.working_volume

    def ok_to_push_out(self, push_out_dist_mm: float) -> bool:
        return push_out_dist_mm <= (
            self.plunger_positions.bottom - self.plunger_positions.blow_out
        )

    def update_nozzle_configuration(
        self,
        back_left_nozzle: str,
        front_right_nozzle: str,
        starting_nozzle: Optional[str] = None,
    ) -> None:
        """
        Update nozzle configuration manager.
        """
        self._nozzle_manager.update_nozzle_configuration(
            back_left_nozzle, front_right_nozzle, starting_nozzle
        )

    def reset_nozzle_configuration(self) -> None:
        """
        Reset nozzle configuration manager.
        """
        self._nozzle_manager.reset_to_default_configuration()

    def add_tip(self, tip_length: float) -> None:
        """
        Add a tip to the pipette for position tracking and validation
        (effectively updates the pipette's critical point)

        :param tip_length: a positive, non-zero float presenting the distance
            in Z from the end of the pipette nozzle to the end of the tip
        :return:
        """
        assert tip_length > 0.0, "tip_length must be greater than 0"
        assert not self.has_tip
        self._has_tip = True
        self._current_tip_length = tip_length

    def remove_tip(self) -> None:
        """
        Remove the tip from the pipette (effectively updates the pipette's
        critical point)
        """
        self._has_tip = False
        self._current_tip_length = 0.0

    @property
    def has_tip(self) -> bool:
        return self._has_tip

    # Cache max is chosen somewhat arbitrarily. With a float is input we don't
    # want this to unbounded.
    @functools.lru_cache(maxsize=100)
    def ul_per_mm(self, ul: float, action: UlPerMmAction) -> float:
        if action == "aspirate":
            fallback = self._active_tip_settings.aspirate.default[
                PIPETTING_FUNCTION_FALLBACK_VERSION
            ]
            sequence = self._active_tip_settings.aspirate.default.get(
                self._pipetting_function_version, fallback
            )
        else:
            fallback = self._active_tip_settings.dispense.default[
                PIPETTING_FUNCTION_FALLBACK_VERSION
            ]
            sequence = self._active_tip_settings.dispense.default.get(
                self._pipetting_function_version, fallback
            )
        return piecewise_volume_conversion(ul, sequence)

    def __str__(self) -> str:
        return "{} current volume {}ul critical point: {} at {}".format(
            self._config.display_name,
            self.current_volume,
            "tip end" if self.has_tip else "nozzle end",
            0,
        )

    def __repr__(self) -> str:
        return "<{}: {} {}>".format(
            self.__class__.__name__, self._config.display_name, id(self)
        )

    def as_dict(self) -> "Pipette.DictType":
        self._config_as_dict.update(
            {
                "current_volume": self.current_volume,
                "available_volume": self.available_volume,
                "name": self.name,
                "model": self.model,
                "pipette_id": self.pipette_id,
                "has_tip": self.has_tip,
                "working_volume": self.working_volume,
                "aspirate_flow_rate": self.aspirate_flow_rate,
                "dispense_flow_rate": self.dispense_flow_rate,
                "blow_out_flow_rate": self.blow_out_flow_rate,
                "default_aspirate_flow_rates": self.aspirate_flow_rates_lookup,
                "default_blow_out_flow_rates": self.blow_out_flow_rates_lookup,
                "default_dispense_flow_rates": self.dispense_flow_rates_lookup,
                "tip_length": self.current_tip_length,
                "return_tip_height": self.active_tip_settings.default_return_tip_height,
                "tip_overlap": self.tip_overlap,
                "back_compat_names": self._config.pipette_backcompat_names,
                "supported_tips": self.liquid_class.supported_tips,
            }
        )
        return self._config_as_dict

    def set_liquid_class_by_name(self, class_name: str) -> None:
        """Change the currently active liquid class."""
        if self.current_volume > 0:
            raise CommandPreconditionViolated(
                "Cannot switch liquid classes when liquid is in the tip"
            )
        if class_name != "default":
            raise InvalidLiquidClassName(
                message=f"Liquid class {class_name} is not valid for {self._config.display_name}"
            )


def _reload_and_check_skip(
    new_config: PipetteConfigurations,
    attached_instr: Pipette,
    pipette_offset: PipetteOffsetByPipetteMount,
    use_old_aspiration_functions: bool,
) -> Tuple[Pipette, bool]:
    # Once we have determined that the new and attached pipettes
    # are similar enough that we might skip, see if the configs
    # match closely enough.
    # Returns a pipette object and True if we may skip hw reconfig
    if (
        new_config == attached_instr.config
        and pipette_offset == attached_instr._pipette_offset
    ):
        # Same config, good enough
        return attached_instr, True
    else:
        newdict = new_config.dict()
        olddict = attached_instr.config.dict()
        changed: Set[str] = set()
        for k in newdict.keys():
            if newdict[k] != olddict[k]:
                changed.add(k)
        if changed.intersection(RECONFIG_KEYS):
            # Something has changed that requires reconfig
            p = Pipette(
                new_config,
                pipette_offset,
                attached_instr._pipette_id,
                use_old_aspiration_functions,
            )
            p.act_as(attached_instr.acting_as)
            return p, False
    # Good to skip
    return attached_instr, True


def load_from_config_and_check_skip(
    config: Optional[PipetteConfigurations],
    attached: Optional[Pipette],
    requested: Optional[PipetteName],
    serial: Optional[str],
    pipette_offset: PipetteOffsetByPipetteMount,
    use_old_aspiration_functions: bool,
) -> Tuple[Optional[Pipette], bool]:
    """
    Given the pipette config for an attached pipette (if any) freshly read
    from disk, and any attached instruments,

    - Compare the new and configured pipette configs
    - Load the new configs if they differ
    - Return a bool indicating whether hardware reconfiguration may be
      skipped
    """

    if not config and not attached:
        # nothing attached now, nothing used to be attached, nothing
        # to reconfigure
        return attached, True

    if config and attached:
        # something was attached and something is attached. are they
        # the same? we can tell by comparing serials
        if serial == attached.pipette_id:
            if requested:
                # if there is an explicit instrument request, in addition
                # to checking if the old and new responses are the same
                # we also have to make sure the old pipette is properly
                # configured to the request
                if requested == str(attached.acting_as):
                    # similar enough to check
                    return _reload_and_check_skip(
                        config, attached, pipette_offset, use_old_aspiration_functions
                    )
            else:
                # if there is no request, make sure that the old pipette
                # did not have backcompat applied
                if str(attached.acting_as) == attached.name:
                    # similar enough to check
                    return _reload_and_check_skip(
                        config, attached, pipette_offset, use_old_aspiration_functions
                    )

    if config:
        return (
            Pipette(config, pipette_offset, serial, use_old_aspiration_functions),
            False,
        )
    else:
        return None, False


def _build_splits(pipette: Pipette) -> Optional[MoveSplit]:
    if pip_types.Quirks.needsUnstick in pipette.config.quirks:
        return MoveSplit(
            split_distance=1,
            split_current=1.75,
            split_speed=1,
            after_time=1800,
            fullstep=True,
        )
    else:
        return None


def generate_hardware_configs(
    pipette: Optional[Pipette], robot_config: RobotConfig, revision: BoardRevision
) -> InstrumentHardwareConfigs:
    """
    Fuse robot and pipette configuration to generate commands to send to
    the motor driver if required
    """
    if pipette:
        return {
            "steps_per_mm": pipette.config.mount_configurations.stepsPerMM,
            "home_pos": pipette.config.mount_configurations.homePosition,
            "max_travel": pipette.config.mount_configurations.travelDistance,
            "idle_current": pipette.plunger_motor_current.idle,
            "splits": _build_splits(pipette),
        }
    else:
        dpcs = robot_config.default_pipette_configs
        return {
            "steps_per_mm": dpcs["stepsPerMM"],
            "home_pos": dpcs["homePosition"],
            "max_travel": dpcs["maxTravel"],
            "idle_current": robot_configs.current_for_revision(
                robot_config.low_current, revision
            )["B"],
            "splits": None,
        }



"""Drop tip in place command request, result, and implementation models."""
from __future__ import annotations
from pydantic import Field, BaseModel
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from .pipetting_common import PipetteIdMixin
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from ..execution import TipHandler


DropTipInPlaceCommandType = Literal["dropTipInPlace"]


class DropTipInPlaceParams(PipetteIdMixin):
    """Payload required to drop a tip in place."""

    homeAfter: Optional[bool] = Field(
        None,
        description=(
            "Whether to home this pipette's plunger after dropping the tip."
            " You should normally leave this unspecified to let the robot choose"
            " a safe default depending on its hardware."
        ),
    )


class DropTipInPlaceResult(BaseModel):
    """Result data from the execution of a DropTipInPlace command."""

    pass


class DropTipInPlaceImplementation(
    AbstractCommandImpl[DropTipInPlaceParams, DropTipInPlaceResult]
):
    """Drop tip in place command implementation."""

    def __init__(
        self,
        tip_handler: TipHandler,
        **kwargs: object,
    ) -> None:
        self._tip_handler = tip_handler

    async def execute(self, params: DropTipInPlaceParams) -> DropTipInPlaceResult:
        """Drop a tip using the requested pipette."""
        await self._tip_handler.drop_tip(
            pipette_id=params.pipetteId, home_after=params.homeAfter
        )

        return DropTipInPlaceResult()


class DropTipInPlace(BaseCommand[DropTipInPlaceParams, DropTipInPlaceResult]):
    """Drop tip in place command model."""

    commandType: DropTipInPlaceCommandType = "dropTipInPlace"
    params: DropTipInPlaceParams
    result: Optional[DropTipInPlaceResult]

    _ImplementationCls: Type[
        DropTipInPlaceImplementation
    ] = DropTipInPlaceImplementation


class DropTipInPlaceCreate(BaseCommandCreate[DropTipInPlaceParams]):
    """Drop tip in place command creation request model."""

    commandType: DropTipInPlaceCommandType = "dropTipInPlace"
    params: DropTipInPlaceParams

    _CommandCls: Type[DropTipInPlace] = DropTipInPlace



"""Wait for resume command request, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from ..execution import RunControlHandler


# NOTE: multiple values accepted for backwards compatibility
# with the 6.0.0-beta.0 release, which used `pause`
WaitForResumeCommandType = Literal["waitForResume", "pause"]


class WaitForResumeParams(BaseModel):
    """Payload required to pause the protocol."""

    message: Optional[str] = Field(
        None,
        description="A user-facing message associated with the pause",
    )


class WaitForResumeResult(BaseModel):
    """Result data from the execution of a WaitForResume command."""


class WaitForResumeImplementation(
    AbstractCommandImpl[WaitForResumeParams, WaitForResumeResult]
):
    """Wait for resume command implementation."""

    def __init__(self, run_control: RunControlHandler, **kwargs: object) -> None:
        self._run_control = run_control

    async def execute(self, params: WaitForResumeParams) -> WaitForResumeResult:
        """Dispatch a PauseAction to the store to pause the protocol."""
        await self._run_control.wait_for_resume()
        return WaitForResumeResult()


class WaitForResume(BaseCommand[WaitForResumeParams, WaitForResumeResult]):
    """Wait for resume command model."""

    commandType: WaitForResumeCommandType = "waitForResume"
    params: WaitForResumeParams
    result: Optional[WaitForResumeResult]

    _ImplementationCls: Type[WaitForResumeImplementation] = WaitForResumeImplementation


class WaitForResumeCreate(BaseCommandCreate[WaitForResumeParams]):
    """Wait for resume command request model."""

    commandType: WaitForResumeCommandType = "waitForResume"
    params: WaitForResumeParams

    _CommandCls: Type[WaitForResume] = WaitForResume
